"""
only for internal use
"""


def get_dict(d, default=None):
    """
     Get the dict instance
    -- If obj is a dict, just return
    -- If obj is not a dict, return default or {}

    :param d:
    :param default:
    :return:
    """
    return d if type(d) == dict else (default or {})


def contains_any(collection, targets):
    """
    Checks whether the collection contains any item

    :param collection:
    :param targets:
    :return:
    """
    collection_type = type(collection)
    if collection_type is dict:
        return any(key in collection for key in targets)
    elif collection_type is list:
        return any(element in collection for element in targets)
    return False


def vie(condition, true_value, false_value):
    """
    If condition is True value return true_value, else return false_value
    False/None/0/''/()/{}/[] is false value

    :param condition:
    :param true_value:
    :param false_value:
    :return:
    """
    return true_value if condition else false_value



import os

from flask import current_app

__all__ = ['init_app_config', 'get_app_config', 'get_app_config_items', 'get_app_path']

app_config = None


def init_app_config(app):
    """
    Initialize application(flask/celery/script/...) configuration.
    Once initialized, the application configuration can be get through get_app_config function anywhere.

    .. versionadded:: 1.5

    Example:
        init_app_config(app)    # Flask
        init_app_config(DevelopmentConfig)  # Class
        init_app_config({...})  # dict


    :param app:Flask/Config/dict
    :return:
    """
    config = get_app_config_items(app)
    if type(config) is dict:
        global app_config
        app_config = config


def get_app_config(key=None, default=None):
    """
    Get the specified config value of the application(flask/celery/script/...).
    If the config is initialized by init_app_config, return the initialized config, otherwise return the config of the current flask application.
    If key is None, return all config values(dict), otherwise return the specified config value

    .. versionupdated::
        1.5 -
        1.6.4 - add default

    Example:
        get_app_config()                        # all Config
        get_app_config('FLASKZ_LOGGER_LEVEL')   # specified config value

    :param key:
    :param default:
    :return:
    """
    config = None
    if app_config:
        config = app_config
    elif current_app:
        config = dict(current_app.config.items())

    if type(config) is not dict:
        config = {}

    if key is None:
        return config
    return config.get(key, default)


def get_app_config_items(app):
    """
    Return config items(dict)

    :param app:
    :return:
    """
    items = None
    if hasattr(app, 'config'):  # Flask
        _app_config = getattr(app, 'config')
        if _app_config:
            items = dict(_app_config.items())
    elif isinstance(app, type):  # Class
        items = cls_to_dict(app)
    else:
        items = app  # dict
    if type(items) is dict:
        return items
    return None


def get_app_path(*path):
    """
    Get the application path.

    Example:
        app.get_app_path("_log", "sys.log")

    :param path:    -The relative path
    :return:        -The path relative to the application root directory
    """
    return os.path.join(os.getcwd(), *path)


from ._cls import cls_to_dict



import json
from collections.abc import Mapping
from datetime import datetime, date, time
from decimal import Decimal

__all__ = [
    'filter_list', 'find_list', 'merge_list', 'each_list', 'get_list',
    'pop_dict_keys', 'del_dict_keys', 'get_deep', 'set_deep', 'merge_dict', 'get_ins_mapping', 'get_dict_mapping', 'get_dict_value_by_type',
    'get_wrap_str', 'is_str', 'str_replace', 'str_contains_any',
    'is_list', 'is_dict', 'slice_str',
    'json_dumps',
    'bulk_append_child',
    'parse_version'
]


# -------------------------------------------list-------------------------------------------
def filter_list(items, func=None, with_index=False, not_none=False):
    """
    Filter the list and return a new list that matches the conditions
    If func is None and not_none is True, return the not None items.

   .. versionupdated::
        1.6.4 - add not_none param

    Example:
        filter_list(items, not_none=True)
        filter_list(items, lambda item: item is not None)
        filter_list(items, lambda index, item: index > 10 and item is not None, True)

    :param items:
    :param func:
    :param with_index:
    :param not_none:
    :return:
    """
    if func is None and not_none is True:  # @2023-11-20 add
        func = _is_not_none

    if with_index is True:
        return _filter_list_with_index(items, func)

    return list(filter(func, items))


def _filter_list_with_index(items, func):
    result = []
    for index, item in enumerate(items):
        if func(item, index) is True:
            result.append(item)
    return result


def find_list(items, func, with_index=False):
    """
    Find an object from the list that matches the conditions

    Example:
        find_list(items, lambda item: item.get('id') == 10)
        find_list(items, lambda index, item: index > 0 and item.get('id') == 10 , True)

    :param with_index:
    :param items:
    :param func:
    :return:
    """
    if with_index is True:
        return _find_list_with_index(items, func)

    for item in items:
        if func(item) is True:
            return item
    return None


def _find_list_with_index(item_list, func):
    for index, item in enumerate(item_list):
        if func(item, index) is True:
            return item
    return None


def merge_list(target_list, *merged_list):
    """
    Merge multiple lists into the first list

    Example:
        nums = [0, 1]
        merge_list(nums, [2, 3], [4, 5]) # nums = [0, 1, 2, 3, 4, 5]

    :param target_list:
    :param merged_list:
    :return:
    """
    for item in merged_list:
        target_list.extend(item)
    return target_list


def each_list(items, func, with_index=True):
    """
    Iterate each object in the list and execute the callback function

    Example:
        each_list(items, lambda item: item['selected'] = True)

    :param with_index:
    :param items:
    :param func:
    :return:
    """
    if with_index is True:
        return _each_list_with_index(items, func)

    for item in items:
        if func(item) is False:
            return


def _each_list_with_index(items, func):
    for index, item in enumerate(items):
        if func(item, index) is False:
            return


def get_list(items, default=None):
    """
    Get the list instance
    -- If obj is a list, just return
    -- If obj is not a list, return default or []
    :param items:
    :param default:
    :return:
    """
    return items if is_list(items) else (default or [])


# -------------------------------------------dict-------------------------------------------
def pop_dict_keys(d, keys):
    """
    Pop the specified keys from the dict object

    Example:
        pop_dict_keys(a_dict, ['name', 'age'])
        pop_dict_keys(a_dict, 'name')

    :param d:
    :param keys:
    :return:
    """
    if type(keys) is not list:
        keys = [keys]
    result = {}
    for key in keys:
        result[key] = d.pop(key, None)
    return result


del_dict_keys = pop_dict_keys


def get_deep(d: dict, key: str, key_split='.', default=None, raising=False):
    """
    Get the deep value of the specified key from the dict object

    Example:
        get_deep(a_dict, 'address.city')
        get_deep(a_dict, 'address/city', key_split='/')


    :param d:
    :param key:
    :param key_split:
    :param default:
    :param raising:
    :return:
    """
    value = d
    try:
        for key in key.split(key_split):
            if isinstance(value, dict):
                value = value[key]
                continue
            else:
                if raising:
                    raise KeyError
                return default
    except KeyError:
        if raising:
            raise
        return default
    else:
        return value


def set_deep(d: dict, key: str, value, key_split='.'):
    """
    Set the deep value of the specified key from the dict object

    Example:
        set_deep(a_dict, 'address.city', 'New York')
        set_deep(a_dict, 'address/city', 'New York', key_split='/')

    :param d:
    :param key:
    :param value:
    :param key_split:
    :return:
    """
    dd = d
    keys = key.split(key_split)
    latest = keys.pop()
    for k in keys:
        dd = dd.setdefault(k, {})
    dd[latest] = value


def get_ins_mapping(ins_list: list, attr: str, deep=False):
    """
    Create a dict map of the specified object list with the specified attribute

    Example:
        get_ins_mapping([User(name='a'), User(name='b', User(name='c')], 'name')  # {'a':user_c, 'b':user_b, 'c':user_c}

    :param ins_list:
    :param attr:
    :param deep:
    :return:
    """

    map_dict = {}
    if deep is True:
        keys = attr.split('.')
        for item in ins_list:
            k_value = item
            for k in keys:
                k_value = getattr(k_value, k, None)
            map_dict[k_value] = item
    else:
        for item in ins_list:
            map_dict[getattr(item, attr, None)] = item

    return map_dict


def get_dict_mapping(dict_list: list, key='id', key_join="+"):
    """
    Create a dict map of the specified dict list withe the specified key

    Example:
        get_dict_mapping([{'name': 'a'}, {'name': 'b'}, {'name': 'c'}], 'name')  # {'a':dict_a, 'b':dict_b, 'c':dict_c}


    :param dict_list:
    :param key:
    :param key_join:
    :return:
    """
    map_dict = {}
    for item in dict_list:
        if isinstance(key, list):
            kv = []
            for k_item in key:
                kv.append(str(get_deep(item, k_item)))
            map_dict[key_join.join(kv)] = item
        else:
            map_dict[get_deep(item, key)] = item

    return map_dict


def merge_dict(d: dict, *merged_dict_list):
    """
    Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.

    .. versionupdated:: 1.5 - return dct

    Example:
        merge_dict(target_dict, a_dict, b_dict)

    :param d: dict onto which the merge is executed
    :param merged_dict_list: dct merged into dct
    :return: dct
    """
    for dict_item in merged_dict_list:
        for k, v in dict_item.items():  # 2022-04-22 dict_item.iteritems-->dict_item.items
            if (k in d and isinstance(d[k], dict)
                    and isinstance(dict_item[k], Mapping)):
                merge_dict(d[k], dict_item[k])
            else:
                d[k] = dict_item[k]
    return d  # @2023-04-12 add result


def get_dict_value_by_type(d: dict, key: str, value_type, default=None, use_isinstance=False):
    """
    Get the dict value of the specified key, if value is not the instance of the specified type, return default value.

    .. versionadded:: 1.2

    Example
        get_dict_value_by_type(a_dict, 'name', str)
        get_dict_value_by_type(a_dict, 'age', int, 10)

    :param d:
    :param key:
    :param value_type:
    :param default:
    :param use_isinstance:
    :return:
    """
    if key not in d:
        return default
    value = d.get(key)
    if use_isinstance is True:
        if value_type is int and (value is True or value is False):  # instance(True,input) == True
            return default
        if isinstance(value, value_type):
            return value
    else:
        if type(value) is value_type:
            return value
    return default


# -------------------------------------------type-------------------------------------------
def is_str(value, use_isinstance=False):
    """
    Check whether the value is string.

    Example:
        is_str('str') # True
        is_str(1) # False

    :param value:
    :param use_isinstance:
    :return:
    """
    if use_isinstance is True:
        return isinstance(value, str)
    return type(value) == str


def is_list(value, use_isinstance=False):
    """
    Check whether the value is list object.

    Example:
        is_list([]) # True
        is_list('abc') # False

    :param value:
    :param use_isinstance:
    :return:
    """
    if use_isinstance is True:
        return isinstance(value, list)
    return type(value) == list


def is_dict(value, use_isinstance=False):
    """
    Check whether the value is dict object.

    Example:
        is_dict({}) # True
        is_list('abc') # False

    :param value:
    :param use_isinstance:
    :return:
    """
    if use_isinstance is True:
        return isinstance(value, dict)
    return type(value) == dict


# -------------------------------------------str-------------------------------------------

def get_wrap_str(*items):
    """
    Use '\n' to join multiple strings.

    :param items:
    :return:
    """
    result = []
    for item in items:
        if item is not None:
            result.append(str(item))
    return '\n'.join(result)


def slice_str(string, start_len=6, end_len=0, ellipsis_str='......'):
    if end_len != 0:
        end_len = -abs(end_len)

    string = str(string)
    if len(string) <= (start_len - end_len):
        return string
    result = string[0:start_len]
    if end_len != 0:
        result = result + ellipsis_str + string[end_len:]
    return result


def str_replace(string, old, new=''):
    """
    Return a copy with all occurrences of substring old replaced by new.

    .. versionadded:: 1.5

    Example:
        str_replace('abca','a','A')             # AbcA
        str_replace('abca',{'a':'A','b':'B'})   # ABcA
        str_replace('abca',['a','b'],"*")       # **c*

    :param string:
    :param old:
    :param new:
    :return:
    """
    v_type = type(old)
    if v_type is dict:
        for key in old:
            string = string.replace(key, old[key])
    elif v_type is list:
        for item in old:
            string = string.replace(item, new)
    else:
        string = string.replace(old, new)
    return string


def str_contains_any(string, substrings):
    """
    Check whether any substring is in the specified string.

    Example:
        str_contains_any('Cisco Nexus Operating System', ['Cisco', 'Nexus']) # True

    :param string: the specified string
    :param substrings: the substring list
    :return:
    """
    return any(sub in string for sub in substrings)


# -------------------------------------------other-------------------------------------------
def bulk_append_child(items, parents, item_parent_key='parent_id', children_key="children", parent_map_key='id'):
    """
    Append the items to the child list of the matched parent object.

    Example:
        bulk_append_child(interface_items,device_list, 'device_id', 'interface_list')

    :param items:
    :param item_parent_key:
    :param parents:
    :param children_key:
    :param parent_map_key:
    :return:
    """
    if not isinstance(items, list):
        items = [items]

    if isinstance(parents, list):
        parents = get_dict_mapping(parents, parent_map_key)

    for item in items:
        parent_attr = item.get(item_parent_key)
        parent_item = parents.get(parent_attr)
        if isinstance(parent_item, dict):
            p_children = parent_item[children_key] = parent_item.get(children_key) or []
            p_children.append(item)


def parse_version(version):
    """
    Parse version string to int list, if version is integer, it will be converted to int

    Example:
        parse_version(sqlalchemy.__version__)   # --> [2, 0, 20] / [2, 0, '0rc1']

    :param version: the version to be parsed, str/int/float/package
    """
    version_type = type(version)

    if version_type is int:  # 1-->1
        return [version]

    if version_type is float:  # 1.2-->[1,2]
        version = str(version)
    elif version_type is not str:  # sqlalchemy,sys
        version = getattr(version, '__version__', None) or getattr(version, 'version', None)

    if type(version) is not str:
        return []

    versions = []
    for v in version.split('.'):
        if v.isdigit():
            versions.append(int(v))
        else:
            versions.append(v)
    return versions


# -------------------------------------------json-------------------------------------------
def json_dumps(obj, **kwargs):
    """
    Serialize obj to a JSON formatted string, with support for additional data types.
    The function extends the default JSON serialization capabilities of Python's json module by
    providing custom serialization for types that are not natively supported,
    such as datetime, date, time, Decimal, bytes, bytearray, set, and custom objects.

    .. versionadded:: 1.8.0

    Example:
        json_dumps({
            'str': 'example',
            'integer': 1,
            'float': 3.14,
            'boolean': True,
            'none': None,
            'list': [1, 2, 3],
            'dict': {'key': 'value'},
            'tuple': (1, 2, 3),

            'set': {1, 2, 3},
            'datetime': datetime.now(),
            'date': date.today(),
            'time': datetime.now().time(),
            'decimal': Decimal('10.5'),
            'bytes': b'byte data',
            'bytearray': bytearray(b'bytearray data')
        }, indent=4)
    :param obj:
    :return:
    """
    if 'default' not in kwargs:
        kwargs['default'] = _json_serializer
    return json.dumps(obj, **kwargs)


def _json_serializer(obj):
    if isinstance(obj, (datetime, date, time)):
        return obj.isoformat()
    if isinstance(obj, (Decimal, bytes, bytearray)):
        return str(obj)
    if isinstance(obj, set):
        return list(obj)
    if hasattr(obj, '__dict__'):
        return obj.__dict__
    raise TypeError(f"Object of type '{obj.__class__.__name__}' is not JSON serializable")


# -------------------------------------------private-------------------------------------------
def _is_not_none(item, *args, **kwargs):
    return item is not None



__all__ = ['ModelRestManager']


class ModelRestManager:
    """
    Used to manage model rest permission and logging.
    """

    def __init__(self):
        self._permission_check = None
        self._login_check = None
        self._logging = None

    def init_app(self, app):
        app.model_rest_manager = self

    def permission_check(self, permission_check):
        self._permission_check = permission_check
        return self._permission_check

    @property
    def permission_check_callback(self):
        return self._permission_check

    def login_check(self, login_check):
        self._login_check = login_check
        return self._login_check

    @property
    def login_check_callback(self):
        return self._login_check

    def logging(self, logging):
        self._logging = logging
        return self._logging

    @property
    def logging_callback(self):
        return self._logging



"""
pip install paramiko
"""
import errno
import os
import re
import socket
import stat
import time
from contextlib import contextmanager
from datetime import datetime
from typing import Optional, Tuple, Union

import paramiko
from paramiko.ssh_exception import SSHException, AuthenticationException


@contextmanager
def ssh_session(hostname: str, username: str, password: str = None, port: Optional[int] = 22, **kwargs):
    """
    SSH contextmanager.

    Example:
        with ssh_session(host, username, password, timeout=20) as ssh:
            ssh.run_command('ls -l')

        with ssh_session(host, username, password, timeout=20) as ssh:
            ssh.run_command_list(['show version', 'show running-config'])

        with ssh_session(host, username, password, timeout=20, secondary_password=enable_pwd, recv_endswith=['# ', '$ ', ': ', '? ', '#']) as ssh:
            ssh.run_command_list(['enable', 'show run'])
    """
    ssh_client = SSH(hostname=hostname, username=username, password=password, port=port, **kwargs)
    yield ssh_client
    ssh_client.close()


def ssh_run_command(connect_kwargs: dict, command: Union[str, list], run_kwargs: Optional[dict] = None) -> Tuple[bool, str]:
    """
    Run the specified command.
    If command is a list, by default, returns the list of results , if run_kwargs.last_result==True returns the result of the last command

    .. versionadded:: 1.6.4
    .. versionupdated::
        1.8.1   - add retries

    Example:
        ssh_run_command({'hostname': hostname, 'username': username, 'password': password},'show running-config')

        ssh_run_command(
            # connect_kwargs
            {'hostname': 'host', 'username': 'username', 'password': 'password',  # host
             'timeout': 10, 'retries': 3,  # connect timeout and retry
             'secondary_password': 'enable_pwd',  # enable password
             'recv_endswith': ['# ', '$ ', ': ', '? ', '#'],  # recv endswith
             'channel_kwargs': {'width': 1000, 'timeout': 2}},  # channel kwargs
            # command
            'show running-config',
            # run_kwargs
            {
                'clean': False,  # not clean output info
                'prompt': False,  # disable prompt
            })

    :param connect_kwargs: the connect kwargs, ex) {'hostname': hostname, 'username': username, 'password': password, 'timeout': 10}
    :param command: the specified command, ex) ls -l
    :param run_kwargs: the run command kwargs, ex) {'recv': False, 'prompt': False}
    :return: (success, result_or_err)
    """
    connect_kwargs = connect_kwargs or {}
    run_kwargs = run_kwargs or {}
    retries = connect_kwargs.pop('retries', None)
    if type(retries) is not int:  # @2025-01-12 add
        retries = 1
    else:
        retries = max(retries, 1)
    retry_count = 0
    while retry_count < retries:
        try:
            with ssh_session(**connect_kwargs) as ssh:
                if type(command) is list:
                    return True, ssh.run_command_list(command, **run_kwargs)
                else:
                    return True, ssh.run_command(command, **run_kwargs)
        except Exception as e:
            if retry_count == retries - 1:
                return False, str(e)
            retry_count = retry_count + 1
            time.sleep(0.1)


def ssh_run_command_list(connect_kwargs: dict, command_list: list, run_kwargs: Optional[dict] = None) -> Tuple[bool, Union[list, str]]:
    """
    Run a command list.
    By default, returns the list of results , if run_kwargs.last_result==True returns the result of the last command

    .. versionadded:: 1.6.4
    .. versionupdated::
        1.8.1   - add retries

    Example:
        ssh.run_command_list({'hostname': hostname, 'username': username, 'password': password}, ['show version', 'show running-config'])

        ssh_run_command_list(
            # connect_kwargs
            {'hostname': 'host', 'username': 'username', 'password': 'password',  # host
             'timeout': 10, 'retries': 3,  # connect timeout and retry
             'secondary_password': 'enable_pwd',  # enable password
             'recv_endswith': ['# ', '$ ', ': ', '? ', '#'],  # recv endswith
             'channel_kwargs': {'width': 1000, 'timeout': 2}},  # channel kwargs
            # command
            ['enable', 'show version'],
            # run_kwargs
            {
                'last_result': True,  # only return last result
                'prompt': False,  # disable prompt
            })

    :param connect_kwargs: the connect kwargs, ex) {'hostname': hostname, 'username': username, 'password': password, 'timeout': 10}
    :param command_list: the command list, ex) ['terminal length 0', 'show version', 'show running-config']
    :param run_kwargs: the run command kwargs, ex) {'recv': False, 'prompt': False}

    :return: (success, result_or_err)
    """
    connect_kwargs = connect_kwargs or {}
    run_kwargs = run_kwargs or {}
    retries = connect_kwargs.pop('retries', None)
    if type(retries) is not int:  # @2025-01-12 add
        retries = 1
    else:
        retries = max(retries, 1)
    retry_count = 0
    while retry_count < retries:
        try:
            with ssh_session(**connect_kwargs) as ssh:
                return True, ssh.run_command_list(command_list, **run_kwargs)
        except Exception as e:
            if retry_count == retries - 1:
                return False, str(e)
            retry_count = retry_count + 1
            time.sleep(0.1)


class SSH(object):
    def __init__(self, hostname, username, password=None, port=22, **kwargs):
        """
        Create a SSH instance.

        .. versionupdated::
            1.6   - add secondary_password and recv_endswith kwargs
            1.6.4 - add connect_kwargs and channel_kwargs kwargs
                  - add prompt param
                  - add prompt param and logic
                  - optimize _recv_data function
            1.7.0 - add transport.is_authenticated() and channel.exit_status_ready() check
                  - add recv_start_delay kwargs
            1.7.3 - add pre_commands kwargs

        Example:
            ssh = SSH(host, username, password)
            ssh.run_command('show version')
            ssh.run_command('ls -l')

            ssh = SSH(host, username, password, timeout=20)
            ssh.run_command_list(['show version', 'show running-config'])
            ssh.run_command_list(['enable', enable_pwd, 'show run'])

            ssh = SSH(host, username, password, timeout=20, secondary_password=enable_pwd, recv_endswith=['# ', '$ ', ': ', '? ', '#'], pre_commands=['terminal length 0'])
            ssh.run_command_list(['enable', 'show run'])
            ssh.run_command_list(['enable', 'show run'], last_result=True) # return last command output
            ssh.run_command_list(['show version', 'show clock', 'show running-config'])

        :param hostname: the host(address) to ssh
        :param username: the username of the host
        :param password: the password of the host
        :param port: the ssh port, default is 22
        :param kwargs:
                - secondary_password: use for enable/sudo action, if None, the password needs to be sent through the command
                - recv_endswith: use for stop receiving, default is ['# ', '$ ', ': ', '? ']
                - pre_commands: the commands that are run before the command/commands are actually run
                - timeout: timeout on blocking read/write operations & socket timer, default is 10
                - connect_kwargs: kwargs for Transport.connect()
        """
        self._username = username
        self._password = password
        self._secondary_password = kwargs.pop('secondary_password', None)  # for enable/sudo
        # recv_endswith = kwargs.pop('recv_endswith', None)  # for stop receiving
        self.recv_endswith = tuple(kwargs.pop('recv_endswith', None) or ['# ', '$ ', ': ', '? '])  # for stop receiving
        self.recv_start_delay = kwargs.pop('recv_start_delay', 0.1)  # @2023-12-31 add, delay before receiving data start
        if 'timeout' in kwargs:
            self._timeout = kwargs.pop('timeout', 0)
        else:
            self._timeout = 10

        if hostname is None and 'host' in kwargs:
            hostname = kwargs.pop('host', None)

        self._pre_commands = None  # @2024-04-14 add
        self._pre_commands_run = False
        pre_commands = kwargs.pop('pre_commands', None)
        if type(pre_commands) is str:
            pre_commands = [pre_commands]
        if type(pre_commands) is list and len(pre_commands) > 0:
            self._pre_commands = pre_commands

        # self.transport = paramiko.Transport((hostname, port))
        self.transport = paramiko.Transport(_create_socket(hostname=hostname, port=port, timeout=self._timeout))
        _connect_kwargs = kwargs.pop('connect_kwargs', None) or {}  # kwargs for Transport.connect()
        self.transport.connect(username=username, password=password, **_connect_kwargs)
        if self.transport.is_authenticated() is not True:  # @2023-12-28 add(Oops, unhandled type 3 ('unimplemented'))
            raise AuthenticationException('Authentication failed.')

        self._channel_kwargs = {'width': 100000, 'height': 100000}
        self._channel_kwargs.update(kwargs.pop('channel_kwargs', {}))
        if 'timeout' in self._channel_kwargs:
            self._channel_timeout = self._channel_kwargs.pop('timeout', 0)
        else:
            self._channel_timeout = self._timeout

        self.ssh = paramiko.SSHClient()
        self.ssh._transport = self.transport
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    @property
    def channel(self):
        if not hasattr(self, '_channel'):
            self._channel = self.ssh.invoke_shell(**self._channel_kwargs)  # @2023-10-31 add width
            if self._channel_timeout > 0:
                self._channel.settimeout(self._channel_timeout)
        return self._channel

    def run_command(self, command: str, recv: bool = True, clean: bool = True, prompt=None) -> Union[str, None]:
        """
        Run the specified command.
        If recv is False, just run the command and return immediately, without waiting for the result

        .. versionupdated::
            1.6.1   - @2023-06-26 add recv parameter
            1.6.3   - @2023-07-10 add clean parameter
            1.6.4   - @2023-10-30 add prompt parameter & logic

        Example:
            ssh.run_command('ls -l')
            ssh.run_command('show run')
            ssh.run_command({
                'command': ' ',
                'clean': False,
            })

         :param command: the command to run.
         :param recv: wait for the result or not.
         :param clean: clean output info or not, default True(clean)
         :param prompt: the prompt of the command line

         :return: the output of the command
        """

        if self._pre_commands_run is False and self._pre_commands:
            self._run_command('', False)
            pre_command_len = len(self._pre_commands)
            for index, pre_command in enumerate(self._pre_commands):
                self._run_command(pre_command, index == pre_command_len - 1)
            self._pre_commands_run = True

        if prompt is None:  # prompt is False for config
            prompt = self.get_prompt()  # 1.get prompt 2.remove welcome

        command, recv, clean = _get_command_arg(command, recv, clean)
        output = self._run_command(command, recv, prompt)
        if output is None:
            return None

        if self._is_secondary_login(command, output):
            secondary_login_output = self._secondary_login(output)
            if secondary_login_output is False:
                return output
            else:
                if _is_enable_command(command, True):  # sudo ls --> secondary_login --> return result, no need to resend
                    output = secondary_login_output
                else:
                    if prompt is not False:
                        prompt = self.get_prompt()
                    output = self._run_command(command, recv)

        if output is None:
            return output

        if clean is not False:
            output = _clean_output(output, command, prompt)

        return output

    def run_command_list(self, command_list: list, last_result: bool = False, **kwargs) -> Union[list, str]:
        """
        Run a command list.
        By default, returns the list of results , if last_result==True returns the result of the last command

        Example:
            ssh.run_command_list(['show version', 'show running-config'])
            ssh.run_command_list(['enable', 'show version'], True)
            ssh.run_command_list(['enable',secondary_password, 'show version'], True)
            ssh.run_command_list(['cd /usr/projects/git/srte',
                      'pwd',
                      'git pull origin master',
                      'wiui@hotmail.com',
                      '11111111'
                      ], True)

         :param command_list: the command list to run.
         :param last_result: if True, return the result of the last command, otherwise, return the result list

         :return: command output result(list/last)
        """
        re_list = []
        prompt = kwargs.get('prompt', None)  # @2024-03-26 kwargs.get('kwargs', None) --> kwargs.get('prompt', None)
        if prompt is None:
            kwargs['prompt'] = self.get_prompt()  # for reuse
        count = len(command_list)
        for index, command in enumerate(command_list):
            if last_result is True:
                if index == count - 1:  # last
                    return self.run_command(command, **kwargs)
                else:
                    self.run_command(command, **kwargs)  # recv can not be false, ex)enable --> recv --> secondary_password
            else:
                re_list.append(self.run_command(command, **kwargs))
        return re_list

    def get_prompt(self) -> str:
        """
        Return the prompt of the command line

        .. versionadded:: 1.6.4

        Example:
            ssh.get_prompt()

        :return:
        """
        output = self._run_command('', recv=True)
        newline_pos = max(output.rfind('\n'), output.rfind('\r'))

        return output[newline_pos + 1:].lstrip('\r\n') if newline_pos != -1 else output

    def close(self):
        """Close the connect"""
        self.channel.close()
        self.ssh.close()

    @property
    def sftp(self):
        if not hasattr(self, '_sftp'):
            self._sftp = paramiko.SFTPClient.from_transport(self.transport)
        return self._sftp

    def sftp_get_dir(self, remote_dir, local_dir):
        """
        Download remote direction to local.
        Return the local files list.

        Example:
            ssh.sftp_get_dir('/usr/projects/git/srte/src/', '/Users/taozh/Work/Codes/ssh_test/sftp')
        """
        if not self._is_path_exists(remote_dir):
            return False
        remote_dir = _remove_end_slash(remote_dir)
        local_dir = _remove_end_slash(local_dir)
        if not os.path.exists(local_dir):
            os.mkdir(local_dir)
        all_files = self.listdir(remote_dir, True)
        local_files = []
        for f in all_files:
            local_filename = f.replace(remote_dir, local_dir)
            local_filepath = os.path.dirname(local_filename)
            local_files.append(local_filepath)
            if not os.path.exists(local_filepath):
                os.makedirs(local_filepath)
            self.sftp.get(f, local_filename)
        return local_files

    def listdir(self, path, recursion=False):
        """
        List all files in the given path.

        Example:
            ssh.listdir('/usr/projects/git/srte/src/')
        """
        all_files = []
        path = _remove_end_slash(path)
        if path[-1] == '/':
            path = path[0:-1]
        files = self.sftp.listdir_attr(path)
        for f in files:
            filename = path + '/' + f.filename
            if stat.S_ISDIR(f.st_mode):  # 如果是文件夹的话递归处理
                if recursion is True:
                    all_files.extend(self.listdir(filename, recursion))
            else:
                all_files.append(filename)
        return all_files

    def _run_command(self, command, recv=True, prompt=None):
        _command = command  # @2025-02-03 remove strip()
        if not _command.strip(' ').endswith(('?', '\t')):  # @2025-02-03 add
            _command = _command + '\n'
        self.channel.send(_command)
        if recv is False:
            return None
        output = self._recv_data(prompt)
        return output

    def _is_secondary_login(self, command, output):
        """
        Determine whether a secondary login is required by checking output text
        """
        if 'assword' not in output:  # input password
            return False
        secondary_password = self._secondary_password  # or self._password
        if secondary_password is None:
            return False
        return _is_enable_command(command)

    def _secondary_login(self, output):
        """
        Secondary login
        return
            - True, if success
            - False, if fail
        """
        secondary_password = self._secondary_password
        pwd_output = self._run_command(secondary_password)
        if pwd_output == output:
            return False
        return pwd_output

    def _is_path_exists(self, path):
        """Return whether the path exists"""
        try:
            self.sftp.stat(path)
        except IOError as e:
            if e.errno == errno.ENOENT:
                return False
            raise
        else:
            return True

    def _recv_data(self, prompt=None):
        """Receive the command output"""
        # cmd_pattern = re.compile('.*[#$] ' + command) # Does not work with password entry

        res_list = []
        if type(prompt) is not str:
            prompt = None

        recv_ready_timeout = self._channel_timeout if self._channel_timeout > 0 else 6
        recv_start = None
        count = 0
        recv_max_bytes = 1024
        recv_start_delay = self.recv_start_delay
        while True:
            if self.channel.exit_status_ready() is True:  # @2023-12-27 add for 'exit' command
                break
            if not self.channel.recv_ready():
                now = datetime.now()
                if recv_start is None:
                    recv_start = now
                else:
                    if (now - recv_start).total_seconds() > recv_ready_timeout:
                        if count == 0:
                            raise socket.timeout()
                        else:
                            break
                time.sleep(0.01)
                continue
            else:
                recv_start = None
                if count == 0:  # Solve the problem of incomplete data
                    time.sleep(recv_start_delay)
            count += 1
            data = self.channel.recv(recv_max_bytes)
            info = data.decode('utf-8', 'backslashreplace')  # @2024-04-18 data.decode() --> current
            res = info.replace(' \r', '')
            res = re.sub(r'\x1b\[[0-9;]*[A-Za-z]', '', res)  # @2025-02-03 add
            res_list.append(res)
            if len(info) < recv_max_bytes:  # read speed > write speed
                if (prompt is not None and info.endswith(prompt)) or info.endswith(self.recv_endswith):
                    time.sleep(0.02)  # Make sure recv is finished, info.endswith(self.recv_endswith) Not 100% sure
                    if not self.channel.recv_ready():
                        break

        return ''.join(res_list)


def _create_socket(hostname, port, timeout=0):
    """
    Create socket instance.
    If timeout>0, set the timeout of the instance, otherwise use the default timeout(maybe 75s).

    :param hostname:
    :param port:
    :param timeout:
    :return:
    """

    reason = 'No suitable address family'
    addrinfos = socket.getaddrinfo(hostname, port, socket.AF_UNSPEC, socket.SOCK_STREAM)
    for family, socktype, proto, canonname, sockaddr in addrinfos:
        if socktype == socket.SOCK_STREAM:
            sock = socket.socket(family, socket.SOCK_STREAM)
            if timeout > 0:
                sock.settimeout(timeout)
            try:
                # retry_on_signal(lambda: sock.connect((hostname, port))) # @2023-06-09 change, remove paramiko.util.retry_on_signal(Paramiko3.0.0 2023-01-20)
                sock.connect((hostname, port))
            except socket.error as e:
                reason = str(e)
            else:
                break
    else:
        raise SSHException(
            'Unable to connect to {}: {}'.format(hostname, reason)
        )
    return sock


def _get_command_arg(command, recv, clean):
    if type(command) is dict:
        _command = command.get('command')  # @2025-02-03 remove strip()
        _recv = command.get('recv') is not False if 'recv' in command else recv
        _clean = command.get('clean') is not False if 'clean' in command else clean
    else:
        _command = command  # @2025-02-03 remove strip()
        _recv = recv
        _clean = clean
    return _command, _recv, _clean


def _is_enable_command(command, sudo=False):
    if sudo is True:
        return command.lower().startswith('sudo ')
    else:
        return command.lower().startswith(('sudo ', 'enable'))


def _clean_output(output, command, prompt):
    """
    Clear the redundant information.
    - Welcome info      ex)Welcome to Ubuntu...
    - Last login info   ex)Last login...
    - Path info         ex)[root@localhost ~]...
    - Command info      ex) [root@localhost ~] ls -l
    - Return prompt     ex) [root@localhost ~]

    """
    # remove the start command  ex) ls -l
    if output.startswith(command):
        output = output[len(command):].lstrip('\r\n')  # 2024-04-05 add '\r\n' to keep space

    prompt_matched = False
    if type(prompt) is str:  # for show
        # 1.remove text before command ex) [root@localhost ~]# ls -l
        cmd_prompt = prompt + command
        cmd_start = output.rfind(cmd_prompt)
        if cmd_start > -1:
            output = output[cmd_start + len(cmd_prompt):]
        prompt_start = output.rfind(prompt)
        # 2.remove the last prompt ex) [root@localhost ~]#
        if prompt_start > -1:
            output = output[:prompt_start]
        prompt_matched = cmd_start > -1 or prompt_start > -1

    if prompt_matched is False:  # for config
        # 1.remove text before command ex) [root@localhost ~]# ls -l
        cmd_pattern = re.compile('.*([#$])( )*' + re.escape(command))  # @2023-06-07 change ([#$])? --> ([#$])
        last_match = None
        for match in cmd_pattern.finditer(output):
            last_match = match
        if last_match:
            output = output[last_match.end():]
        # 2.remove the last prompt ex) [root@localhost ~]#
        path_pattern = re.compile('.*[#$]( )?$')  # @2023-07-06 change '.*[#$] ' --> '.*[#$]( )?'
        path_match = path_pattern.search(output)  # @2023-10-26 change sub()-->search(), '.*[#$]( )?'-->'.*[#$]( )?$'
        if path_match:
            path_match_start, path_match_end = path_match.span()
            output = output[:path_match_start] + '' + output[path_match_end:]
        output = output.replace(command + '\r\n', '')

    return output.strip('\r\n')  # 2024-04-05 add '\r\n' to keep space


def _remove_end_slash(path):
    """Remove ending slash """
    if path[-1] == '/':
        return path[0:-1]
    return path



from ._app import *
from ._cache import *
from ._cls import *
from ._common import *
from ._func import *
#
from ._magic import *
#
from ._request_api import *
#
from ._request_args import *
from ._response import *
from ._timer import *



"""
pip install pycryptodome
"""
import base64
import random
import string

from Crypto.Cipher import PKCS1_OAEP, AES
from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5
from Crypto.Util.Padding import pad, unpad

__all__ = ['RSACipher', 'AESCipher']


class RSACipher:
    """
    The rsa utility class.

    The result of the encrypt and sign methods are in base64 format
    """

    @staticmethod
    def generate_key(bits=1024, to_base64=True):
        """
        Generate ras key(private_key, public_key)

        :param bits:
        :param to_base64: if not False, return the key with base64 format

        :return: (private_key, public_key)
        """
        rsa_key = RSA.generate(bits)
        private_key = rsa_key.export_key('PEM').decode('utf-8')
        public_key = rsa_key.public_key().export_key('PEM').decode('utf-8')
        if to_base64 is not False:
            private_key = base64.b64encode(private_key.encode('utf-8')).decode('utf-8')
            public_key = base64.b64encode(public_key.encode('utf-8')).decode('utf-8')
        return private_key, public_key

    @staticmethod
    def encrypt(plaintext, public_key):
        """
        Encrypt data with rsa public key.

        Example:
            encrypt_key = RSACipher.encrypt(public_key, aes_key)

        :param plaintext: the data to be encrypted
        :param public_key: rsa public key

        :return: ciphertext：the encrypted text(base64 format)
        """
        plaintext = _to_byte(plaintext)
        cipher = PKCS1_OAEP.new(RSACipher._import_rsa_key(public_key), hashAlgo=SHA256)
        return base64.b64encode(cipher.encrypt(plaintext)).decode('utf-8')

    @staticmethod
    def decrypt(ciphertext, private_key):
        """
        Decrypt ciphertext with rsa private key.

        Example:
            aes_key = RSACipher.decrypt(private_key, encrypt_key)

        :param ciphertext: the data to be decrypted(base64 format)
        :param private_key: rsa private key

        :return: plaintext: the decrypted string
        """
        ciphertext = _to_byte(ciphertext)
        cipher = PKCS1_OAEP.new(RSACipher._import_rsa_key(private_key), hashAlgo=SHA256)
        return cipher.decrypt(base64.b64decode(ciphertext)).decode('utf-8')

    @staticmethod
    def sign(text, private_key):
        """
        Sign the data with the private key.

        Example:
            signature = RSACipher.sign(private_key, aes_key)

        :param text: the text to be signed
        :param private_key: rsa private key

        :return: signature: the signature(base64 format)
        """
        text = _to_byte(text)
        signer = PKCS1_v1_5.new(RSACipher._import_rsa_key(private_key))
        signature = signer.sign(SHA256.new(text))
        return base64.b64encode(signature).decode('utf-8')

    @staticmethod
    def verify(text, signature, public_key):
        """
        Verify the signature.

        Example:
            result = RSACipher.verify(public_key, aes_key, signature)

        :param text: the raw text
        :param signature: the signature(base64 format)
        :param public_key: rsa public key

        :return: result: return True on success, otherwise return False
        """
        text = _to_byte(text)
        try:
            verifier = PKCS1_v1_5.new(RSACipher._import_rsa_key(public_key))
            return verifier.verify(SHA256.new(text), base64.b64decode(signature))
        except (Exception,):
            return False

    @staticmethod
    def _import_rsa_key(key):
        if _is_base64(key):
            key = base64.b64decode(key)
        return RSA.importKey(key)


class AESCipher:
    """
    The aes utility class.

    The result of the encrypt is in base64 format.
    """

    @staticmethod
    def generate_key():
        """
        Generate a 16-bit random key

        Example:
            aes_key = AESCipher.generate_key()

        :return: key
        """
        key = ''.join(random.choice(string.digits + string.ascii_letters) for _ in range(16))
        return key

    @staticmethod
    def encrypt(plaintext, key, iv=None):
        """
        Encrypt data with aes key and iv.

        Example:
            encrypt_text = AESCipher.encrypt(data, aes_key)

        :param plaintext: the data to be encrypted
        :param key: the key used to create AES cipher
        :param iv: the iv used to create AES cipher(AES.MODE_CBC), if None, use key as iv.

        :return: ciphertext: the encrypted text(base64 format)
        """
        plaintext = _to_byte(plaintext)
        plaintext = pad(plaintext, AES.block_size)
        key, iv = AESCipher._encode_key_iv(key, iv)
        cipher = AES.new(key, AES.MODE_CBC, iv)
        return base64.b64encode(cipher.encrypt(plaintext)).decode('utf-8')

    @staticmethod
    def decrypt(ciphertext, key, iv=None):
        """
        Decrypt ciphertext with aes key and iv.

        Example:
            decrypt_text = AESCipher.decrypt(encrypt_text, aes_key)

        :param ciphertext: the ciphertext to be decrypted(base64 format)
        :param key: the key used to decrypt
        :param iv: the iv used to decrypt, if None, use key as iv.

        :return: plaintext: the decrypted string
        """
        ciphertext = _to_byte(ciphertext)
        ciphertext = base64.b64decode(ciphertext)
        key, iv = AESCipher._encode_key_iv(key, iv)
        cipher = AES.new(key, AES.MODE_CBC, iv)
        decrypted_data = cipher.decrypt(ciphertext)
        decrypted_data = unpad(decrypted_data, AES.block_size)
        decrypted_data = decrypted_data.decode('utf-8')
        return decrypted_data

    @staticmethod
    def _encode_key_iv(key, iv=None):
        if type(key) is str:
            key = key.encode('utf-8')
        if iv is None:
            iv = key
        elif type(iv) is str:
            iv = iv.encode('utf-8')

        return key, iv


def _is_base64(txt):
    try:
        return base64.b64encode(base64.b64decode(txt)).decode('utf-8') == txt.replace('\n', '').replace(' ', '').strip()
    except Exception:
        return False


def _to_byte(text):
    if type(text) is str:
        return text.encode('utf-8')
    return text


if __name__ == '__main__':
    """
    RSA+AES
    The process of sending encrypted information from the client to the server
    1.The client encrypts the data with the server's public key
    2.The client signs the aes key with the client's private key
    3.The client sends ciphertext and signature to server
    4.The server uses the client's public key to verify the signature
    5.The server uses the server's private key to decrypt the ciphertext
    """
    server_private_key, server_public_key = RSACipher.generate_key()  # just for test
    client_private_key, client_public_key = RSACipher.generate_key()  # just for test

    # --------------------client--------------------
    send_data = 'hello server'
    # Generate aes key
    send_aes_key = AESCipher.generate_key()
    print('--AES Key:%s' % send_aes_key)

    # encrypt [data] with aes key
    send_aes_encrypted_data = AESCipher.encrypt(send_data, send_aes_key)
    print('--Encrypted Data: %s' % send_aes_encrypted_data)

    # sign [aes key] with client rsa private key
    send_aes_key_signature = RSACipher.sign(send_aes_key, client_private_key)
    print('--AES Key Signature: %s' % send_aes_key_signature)

    # encrypt [aes key] with server rsa public key
    send_encrypted_aes_key = RSACipher.encrypt(send_aes_key, server_public_key)
    print('--Encrypted AES Key: %s' % send_encrypted_aes_key)

    # send (send_aes_key_signature, send_encrypted_aes_key, send_aes_encrypted_data)
    print('--send (send_aes_key_signature, send_encrypted_aes_key, send_aes_encrypted_data)')
    print('\n************************************************\n')
    print('--receive (send_aes_key_signature, send_encrypted_aes_key, send_aes_encrypted_data)')
    # receive (send_aes_key_signature, send_encrypted_aes_key, send_aes_encrypted_data)

    # --------------------server--------------------
    # decrypt [aes key] with server rsa private key
    rcv_aes_key = RSACipher.decrypt(send_encrypted_aes_key, server_private_key)
    print('--Decrypted AES Key: %s' % rcv_aes_key)

    # verify aes_key and signature with client rsa public key
    result = RSACipher.verify(rcv_aes_key, send_aes_key_signature, client_public_key)
    print('--Verify AES Key and AES Key Signature: %s' % result)

    # decrypt [data] with [aes key]
    rcv_decrypted_data = AESCipher.decrypt(send_aes_encrypted_data, rcv_aes_key)
    print('--Decypted Data: %s' % rcv_decrypted_data)



from inspect import isfunction

__all__ = ['Attribute', 'cls_to_dict', 'ins_to_dict']


class Attribute:
    """
    Used to create attribute object

    Example:
        obj = AttrCls(**{'a':1,'b':2})
        obj.c = 3
        obj.a = 10

        print(obj.x)
        print(obj)
    """

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getattr__(self, attr):
        return None

    def __repr__(self):
        t = []
        for key, value in self.__dict__.items():
            t.append(key + '=' + str(value))
        return 'Attribute(' + ", ".join(t) + ')'


def cls_to_dict(cls, option=None):  # @2023-03-15 add, convert class(config) to dictionary
    """
    Convert class(config) to dictionary.

     .. versionadded:: 1.5

     Example:
         cls_to_dict(Config)
         cls_to_dict(TestConfig)

    :param cls:
    :param option:
    :return:
    """
    if option is None:
        option = {}
    attr_filter = _get_option_filter(option)

    props = {}
    cls_dir = dir(cls.__class__)
    for key in dir(cls):
        if key.startswith('__') or key in cls_dir:
            continue
        if _filter_attr(cls, key, attr_filter) is False:
            continue

        value = getattr(cls, key)
        if callable(value):
            continue
        props[key] = value

    return props


def ins_to_dict(ins, option=None):
    """
    Convert instance object to dictionary.

    Example:
    ins_to_dict(A, {
        'cascade': 3,  # 如果子项没有cascade，则使用父项-1，如果cascade不大于0，则不对象属性
        'recursion_value': '{...}'
        # 'include': ['a1'],  # 只有include中的字段才会返回，不区分值是否是对象，include的优先级>exclude
        # 'exclude': ['a2'],  # 只有exclude外的字段才会返回，不区分值是否是对象
        # 'getattrs': lambda ins, item_cascade, option, path_keys: ins.__dict__.items(),
        #
        # 'bb': {  # 某个子项的设置
        #     # 'cascade': 1,
        #     # 'include': ['b1'],
        #     'exclude': ['b2'],
        # },
        # 'bb.xx': {  # 子项的子项的设置
        #     'exclude': ['x1']
        # },
        # 'bb.yy': {
        #     'exclude': ['y2']
        # },
        # 'cc': {  # cc列表中的每个数据的设置
        #     'exclude': ['n'],
        #     'cascade': 1
        # },
        # "cc.xx": {
        #     'exclude': ['x2']
        # }
    })

    :param ins:
    :param option:
    :return:
    """
    if isinstance(ins, list):
        result = []
        for item in ins:
            result.append(_ins_to_dict(item, option))
    else:
        result = _ins_to_dict(ins, option)
    return result


def _ins_to_dict(ins, option=None, path_items=None, path_keys=None):
    if path_items is None:
        path_items = []
    else:
        path_items = path_items.copy()
    path_items.append(ins)

    if path_keys is None:
        path_keys = []

    if option is None:
        option = {}

    if len(path_keys) > 0:
        prefix = '.'.join(path_keys)
        item_option = option.get(prefix, {})
    else:
        item_option = option

    if 'cascade' in item_option:
        item_cascade = item_option.get('cascade', 0)
    else:
        item_cascade = _get_option_key(option, path_keys, 'cascade')  # print('查找父option的cascade，直到找到根option')
        if type(item_cascade) == int:
            item_cascade -= len(path_keys)

    if type(item_cascade) is not int:
        item_cascade = 0

    item_relationships = item_option.get('relationships', None)  # @2024-04-14 add
    if type(item_relationships) is not list:
        item_relationships = []

    item_recursion_value = _get_option_key(option, path_keys, 'recursion_value')

    item_getattrs = _get_option_key(option, path_keys, 'getattrs')
    if item_getattrs:
        attrs = item_getattrs(ins, item_cascade, item_relationships, option, path_keys)
    else:
        attrs = ins.__dict__

    # with_none_value = item_option.get('with_none_value', False)
    attr_filter = _get_option_filter(item_option)
    result = {}
    for key, value in attrs.items():
        if _filter_attr(ins, key, attr_filter) is False:
            continue

        _value = None
        if isinstance(value, list):
            _value = []
            _isinstance = False
            for item in value:
                if hasattr(item, '__dict__'):
                    _isinstance = True
                    if item_cascade > 0:
                        if item not in path_items:
                            _value.append(_ins_to_dict(item, option, path_items, path_keys + [key]))
                        elif item_recursion_value is not None:
                            _value.append(item_recursion_value)
                else:
                    _value.append(item)
            if _isinstance is True and len(_value) == 0:
                _value = None
        elif hasattr(value, '__dict__'):
            if item_cascade > 0:
                if value not in path_items:
                    _value = _ins_to_dict(value, option, path_items, path_keys + [key])
                elif item_recursion_value is not None:
                    _value = item_recursion_value
        else:
            _value = value

        # if _value is not None:  # @2024-03-06 remove
        result[key] = _value
        # elif with_none_value:
        #     result[key] = None
    return result


def _get_option_key(option, key_list, key):
    """
    Return the specified value from the option, if not found, search from its parent option until it is found
    """
    if len(key_list) == 0:
        return option.get(key)

    item_key = '.'.join(key_list)
    item_option = option.get(item_key)
    if isinstance(item_option, dict):
        if key in item_option:
            return item_option.get(key)
    return _get_option_key(option, key_list[:-1], key)


def _get_option_filter(option):
    include = option.get('include', [])
    exclude = option.get('exclude', [])

    if isfunction(include):
        has_include = 'func'
    else:
        has_include = len(include) > 0

    if isfunction(exclude):
        has_exclude = 'func'
    else:
        has_exclude = len(exclude) > 0
    return {
        'include': include,
        'has_include': has_include,
        'exclude': exclude,
        'has_exclude': has_exclude,
    }


def _filter_attr(obj, attr, attr_filter):
    has_include = attr_filter.get('has_include')
    include = attr_filter.get('include')
    has_exclude = attr_filter.get('has_exclude')
    exclude = attr_filter.get('exclude')

    if has_include is True and attr not in include:
        return False
    elif has_include == "func" and include(obj, attr) is not True:
        return False

    if has_exclude is True and attr in exclude:
        return False
    elif has_exclude == "func" and exclude(obj, attr) is True:
        return False

    return True



from sqlalchemy import desc, asc, Column, and_, or_
from sqlalchemy.sql.elements import UnaryExpression

__all__ = ['parse_pss']


def parse_pss(cls, pss_payload=None):
    """
    Get the search, sort and paging information.
    flaskz.utils.get_pss == flaskz.models.parse_pss
    pss = page+search+sort

    .. versionadded:: 1.6.1 - rename flaskz.utils.get_pss function to flaskz.models.parse_pss, flaskz.utils.get_pss is still available
    .. versionupdated::
        - 1.6.1: change return list item(avoid SQL injection): SQL text --> Column.operator(parameter)  ex) "name like '%admin%'" --> TemplateModel.name.like('%admin%')
        - 1.7.0: add relationship-related search and sort parameter parsing
        - 1.7.2: add like_columns parameter parsing

    Example:
        result, result = TemplateModel.query_pss(parse_pss(   # use flaskz.models.parse_pss to parse pss payload
            TemplateModel, {   # FROM templates
                "search": {                         # WHERE
                    "like": "t",                    # name like '%t%' OR description like '%t%' (TemplateModel.like_columns = ['name', description])
                    "age": {                        # AND (age>1 AND age<20)
                        ">": 1,                     # operator:value, operators)'='/'>'/'<'/'>='/'<='/'BETWEEN'/'LIKE'/'IN'
                        "<": 20
                    },
                   "city": {                        # AND (city IN ('Paris','London'))
                        "in": ['Paris' ,'London'],
                    },
                    # "address.city": "New York",   # *relation
                    # "address": {                  # *relation like
                    #     "like": True,
                    #     "like_columns": ["city"]  # like columns of the relation
                    # },
                    "email": "taozh@focus-ui.com",  # AND (email='taozh@focus-ui.com')
                    "_ors": {                       # AND (country='America' OR country='Canada')
                        "country": "America||Canada"
                    },
                    "_ands": {                      # AND (grade>1 AND grade<5)
                        "grade": {
                            ">": 1,
                            "<": 5
                        }
                    }
                },
                "sort": {                           # ORDER BY templates.name ASC
                    "field": "name",
                    "order": "asc"
                },
                # "sort":[                            # multiple sort
                #     {"field": "name", "order": "asc"},
                #     {field": "age", "order": "desc"}
                # ],
                "page": {                           # LIMIT ? OFFSET ? (20, 0)
                    "offset": 0,
                    "size": 20
                }
            }))

    :param cls: Model class
    :param pss_payload: page+search+sort
    :return:
    """
    pss_payload = get_dict(pss_payload)
    # --------------------search--------------------
    search = pss_payload.get('search') or pss_payload.get('query') or {}
    parse_option = _get_parse_options(cls, pss_payload)
    pss_options = {}
    pss_options.update(_parse_search_filters(cls, search, parse_option))
    relationships_pss = _parse_relationships_search_filters(cls, pss_payload, pss_options, parse_option)  # @2023-12-05 add relationship search
    _merge_search_filter_likes(pss_options, parse_option)
    # distinct = []  # @2023-06-07 add
    # _distinct = search.pop('_distinct', None)
    # if _distinct:
    #     if is_str(_distinct):
    #         _distinct = [_distinct]
    #
    #     if is_list(_distinct):
    #         for distinct_item in _distinct:
    #             distinct_column = cls.get_column_by_field(distinct_item)
    #             if distinct_column is not None:
    #                 distinct.append(distinct_column)

    # --------------------group--------------------
    groups = _parse_groups(cls, pss_payload.get('group', None))  # @2023-06-07 add

    # --------------------page--------------------
    page = pss_payload.get('page') or {}
    offset = page.get('offset') or page.get('skip') or 0
    limit = page.get('limit') or page.get('size') or 100000  # @2023-11-22 remove 'or 100000'

    # --------------------sort--------------------
    # @2022-04-10 fix exception subs2 = relationship('PerfTestSubModel2', cascade='all,delete-orphan', lazy='joined') ->
    # Can't resolve label reference for ORDER BY / GROUP BY / DISTINCT etc. Textual SQL expression 'f2' should be explicitly declared as text('f2')
    orders = _parse_sorts(cls, pss_payload.get('sort'))

    pss_options.update({
        'order': _filter_pss_list(orders),
        'group': _filter_pss_list(groups),

        'offset': offset,
        'limit': limit,

        'relationships': relationships_pss
        # 'distinct': distinct
    })
    return pss_options


# -------------------------------------------search-------------------------------------------
def _parse_search_filters(cls, search, parse_option):
    """
    Get the search options
    """
    # --------------------search--------------------
    # search = search or {}
    # @2023-12-05 add _like & _ilike

    like_filters, ilike_filters, notlike_filters, notilike_filters = _parse_search_like_filters(cls, search, *_get_search_like_values(cls, search))
    ands = []
    ors = []

    search_ands_field, search_ands = _get_first_non_column_field_value(cls, search, _get_parse_option_keywords('ands'))  # search.pop('_ands', None)
    if search_ands:
        for key in search_ands:
            _append_search_filters(ands, cls, key, search_ands[key], parse_option)
    search_ors_field, search_ors = _get_first_non_column_field_value(cls, search, _get_parse_option_keywords('ors'))  # search.pop('_ors', None)
    if search_ors:
        for key in search_ors:
            _append_search_filters(ors, cls, key, search_ors[key], parse_option)

    for key in search:
        _append_search_filters(ands, cls, key, search[key], parse_option)

    return {
        'filter_ands': _filter_pss_list(ands),
        'filter_ors': _filter_pss_list(ors),

        'filter_likes': _filter_pss_list(like_filters),
        'filter_ilikes': _filter_pss_list(ilike_filters),
        'filter_notlikes': _filter_pss_list(notlike_filters),
        'filter_notilikes': _filter_pss_list(notilike_filters),
    }


def _parse_relationships_search_filters(cls, pss_payload, pss_options, parse_option):
    """
    Returns pss config of the relationships

    .. versionupdated::
        - 1.8.2: change the key of relationships_filters to InstrumentedAttribute

    :param cls:
    :param pss_payload:
    :return:
    """
    # @2023-12-05 add
    relationships = cls.get_relationships()
    relationships_filters = {}
    search = pss_payload.get('search') or {}

    relationships_search = {}  # Role:{'name':'admin'}
    for key, value in search.items():  # "role.name":"admin"
        col = cls.get_column_by_field(key)
        if col is not None:
            continue
        keys = key.split('.')
        if len(keys) == 2:
            relationship_key = keys[0]
            relationship_filed = keys[1]
            relationship = relationships.get(relationship_key)
            if relationship:
                relationships_search.setdefault(relationship, {}).update({relationship_filed: value})
    for key in relationships.keys():  # "role":{"name":"admin","like":"administrator"}
        col = cls.get_column_by_field(key)  # ?
        if col is not None:
            continue
        relationship_search_value = search.get(key, None)
        if type(relationship_search_value) is dict:
            relationship = relationships.get(key)
            relationships_search.setdefault(relationship, {}).update(relationship_search_value)

    cls_search_like, cls_search_ilike, cls_search_notlike, cls_search_notilike = _get_search_like_values(cls, search)
    for relationship, relationship_search_payload in relationships_search.items():
        # @2023-12-12 add
        # Merge into global like query
        relationship_cls = relationship.mapper.class_
        relationship_search = dict(relationship_search_payload)
        (search_like_field, search_like), (
            search_ilike_field, search_ilike), (
            search_notlike_field, search_notlike), (
            search_notilike_field, search_notilike) = _get_search_like_values(relationship_cls, relationship_search, True)
        if search_like is True:
            relationship_search[search_like_field] = cls_search_like
        if search_ilike is True:
            relationship_search[search_ilike_field] = cls_search_ilike
        if search_notlike is True:
            relationship_search[search_notlike_field] = cls_search_notlike
        if search_notilike is True:
            relationship_search[search_notilike_field] = cls_search_notilike

        relationship_pss_options = _parse_search_filters(relationship_cls, relationship_search, parse_option)
        relationships_filters[getattr(cls, relationship.key)] = relationship_pss_options  # @2025-06-18 RelationshipProperty --> InstrumentedAttribute

        if search_like is True:
            pss_options.setdefault('filter_likes', []).extend(relationship_pss_options.pop('filter_likes', []))
        if search_ilike is True:
            pss_options.setdefault('filter_ilikes', []).extend(relationship_pss_options.pop('filter_ilikes', []))
        if search_notlike is True:
            pss_options.setdefault('filter_notlikes', []).extend(relationship_pss_options.pop('filter_notlikes', []))
        if search_notilike is True:
            pss_options.setdefault('filter_notilikes', []).extend(relationship_pss_options.pop('filter_notilikes', []))

        _merge_search_filter_likes(relationship_pss_options, parse_option)

    return relationships_filters


def _append_search_filters(items, cls, key, value, parse_option):
    """
    Append filter item
    """
    col = cls.get_column_by_field(key)
    if col is None:
        return items
    ignore_null = parse_option.get('ignore_null')
    str_sep = parse_option.get('str_sep')
    if value is not None or ignore_null is False:
        if is_str(value):
            # value = value.strip()  # @2023-12-12 remove
            if value != '' or ignore_null is False:
                if str_sep and value != str_sep:
                    val_arr = value.split(str_sep)
                    for op_v in val_arr:
                        items.append(get_col_op(col, '==', op_v))
                else:
                    items.append(get_col_op(col, '==', value))
        elif is_dict(value):
            for operator, op_v in value.items():
                items.append(get_col_op(col, operator, op_v))
        else:
            items.append(get_col_op(col, '==', value))
    return items


def get_col_op(column, operator, value):
    """
        < , <= , == , != , > , >=
        in , notin , between ,
        is , isnot
        like , ilike , notlike , notilike

        ~startswith , endswith , contains~
    """
    operator = operator.lower()
    op_func_mapping = {
        'in': lambda col, val: col.in_(val) if type(val) is list else None,
        'notin': lambda col, val: col.notin_(val) if type(val) is list else None,
        'between': lambda col, val: col.between(*val) if type(val) is list else None,

        'is': lambda col, val: col.is_(val),
        'isnot': lambda col, val: col.is_not(val),

        'like': lambda col, val: col.like(val),
        'ilike': lambda col, val: col.ilike(val),
        'notlike': lambda col, val: col.notlike(val),
        'notilike': lambda col, val: col.notilike(val),

        '==': lambda col, val: col.__eq__(val)
    }
    value = _get_operator_value(operator, value)  # 2024-04-07 add
    if operator in op_func_mapping:
        return op_func_mapping.get(operator)(column, value)
    return column.op(operator)(value)


def _get_parse_options(cls, pss_payload):
    search = pss_payload.get('search') or pss_payload.get('query') or {}
    return {
        'str_sep': _get_parse_option(cls, search, _get_parse_option_keywords('str_sep'), '||', '||'),
        'ignore_null': _get_parse_option(cls, search, _get_parse_option_keywords('ignore_null'), True) is not False,  # 2023-12-12 add
        'like_join': 'and' if _get_parse_option(cls, search, _get_parse_option_keywords('like_join'), 'or', 'or').lower() == 'and' else 'or',  # 2023-12-23 add
        'notlike_join': 'or' if _get_parse_option(cls, search, _get_parse_option_keywords('notlike_join'), 'and', 'and').lower() == 'or' else 'and'
    }


def _get_parse_option(cls, search_payload, keys, not_contained_default, none_default=None):
    """
    Get pss_payload parsing option item
    """
    if not contains_any(search_payload, keys):
        return not_contained_default
    opt = _get_first_non_column_field_value(cls, search_payload, keys)[1]
    if opt is None:
        return none_default
    return opt


def _get_first_non_column_field_value(cls, props, fields):
    """
    Get the value of the first non-column field in the specified dictionary
    """
    for field in fields:
        if field in props and cls.get_column_by_field(field) is None:
            return field, props.get(field)
    return None, None


def _filter_pss_list(items):
    return [item for item in items if item is not None]


def _get_parse_option_keywords(keyword):
    return ['_' + keyword, keyword]


# -------------------------------------------search/like-------------------------------------------
def _parse_search_like_filters(cls, search, search_like, search_ilike, search_notlike, search_notilike):
    """
    Get like list by like_columns.
    """

    like_columns = []
    cls_like_columns_field, cls_like_columns = _get_first_non_column_field_value(cls, search, _get_parse_option_keywords('like_columns'))  # @2024-01-17 add
    if type(cls_like_columns) is not list:
        cls_like_columns = getattr(cls, 'like_columns', [])
    for col in cls_like_columns:
        if is_str(col):
            col = cls.get_column_by_field(col)
        if col is not None:
            like_columns.append(col)
    if len(like_columns) == 0:
        return [], [], [], []

    like_filters = _gen_like_columns_filters(like_columns, 'like', search_like)
    ilike_filters = _gen_like_columns_filters(like_columns, 'ilike', search_ilike)
    notlike_filters = _gen_like_columns_filters(like_columns, 'notlike', search_notlike)
    notilike_filters = _gen_like_columns_filters(like_columns, 'notilike', search_notilike)
    return like_filters, ilike_filters, notlike_filters, notilike_filters


def _get_search_like_values(cls, search, include_field=False):
    search_like_field, search_like = _get_first_non_column_field_value(cls, search, _get_parse_option_keywords('like'))
    search_ilike_field, search_ilike = _get_first_non_column_field_value(cls, search, _get_parse_option_keywords('ilike'))
    search_notlike_field, search_notlike = _get_first_non_column_field_value(cls, search, _get_parse_option_keywords('notlike'))
    search_notilike_field, search_notilike = _get_first_non_column_field_value(cls, search, _get_parse_option_keywords('notilike'))
    if include_field is True:
        return [(search_like_field, search_like), (search_ilike_field, search_ilike), (search_notlike_field, search_notlike), (search_notilike_field, search_notilike)]
    return [search_like, search_ilike, search_notlike, search_notilike]


def _gen_like_columns_filters(like_columns, like_op, like_value):
    """Generate the like filters"""
    if like_value is None or type(like_value) is not str or like_value == '':
        return []

    # if not (like_value.startswith('%') or like_value.startswith('%')):
    #     like_value = "%" + like_value + "%"
    like_value = _get_operator_value(like_op, like_value)
    filters = []
    for col in like_columns:
        # filters.append(get_col_op(col, like_op, like_value))
        filters.append(get_col_op(col, like_op, like_value))
    return filters


def _merge_search_filter_likes(pss_options, parse_option):
    filters = []
    filter_likes = pss_options.pop('filter_likes', [])
    filter_likes.extend(pss_options.pop('filter_ilikes', []))
    if len(filter_likes) > 0:
        if parse_option.get('like_join') == 'and':
            filters.append(and_(*filter_likes))
        else:
            filters.append(or_(*filter_likes))

    filter_notlikes = pss_options.pop('filter_notlikes', [])
    filter_notlikes.extend(pss_options.pop('filter_notilikes', []))
    if len(filter_notlikes) > 0:
        if parse_option.get('notlike_join') == 'or':
            filters.append(or_(*filter_notlikes))
        else:
            filters.append(and_(*filter_notlikes))
    pss_options['filter_likes'] = filters


def _get_operator_value(operator, value):
    if operator in ('like', 'ilike', 'notlike', 'notilike'):
        if not (value.startswith('%') or value.endswith('%')):
            return "%" + value + "%"
    return value


# -------------------------------------------sort+group-------------------------------------------
def _parse_groups(cls, group):
    """
    Get group by list
    """
    groups = []
    if group:
        if is_str(group):
            group = [group]
        if is_list(group):
            for group_item in group:
                group_column = cls.get_column_by_field(group_item)
                if group_column is not None:
                    groups.append(group_column)
    return groups


def _parse_sorts(cls, sort):
    """
    Get sort list.
    @2023-12-05 add relationship sort
    """
    orders = []
    if is_list(sort):
        sorts = sort
    else:  # @2023-08-31 elif is_dict(sort):-->else
        sorts = [sort]

    for sort_item in sorts:
        if sort_item is None or sort_item == '':
            continue

        if is_str(sort_item):  # sort: "name"/["name"...]  'role.name'
            sort_column = _get_column_by_field(cls, sort_item)
            if sort_column is not None:
                orders.append(asc(sort_column))
        elif type(sort_item) is Column:  # User.name
            if sort_item in cls.get_columns():
                orders.append(asc(sort_item))
        elif type(sort_item) is UnaryExpression:  # desc(User.name)/asc(User.name)
            orders.append(sort_item)
        elif is_dict(sort_item):  # {"field": "name", "order": "desc"}
            _order = sort_item.get('order')
            if _order and not is_str(_order):  # {"order": desc(User.name))} == "sort":desc(User.name)
                orders.append(_order)
            else:
                sort_field = sort_item.get('field')  # {"field": "name", "order": "asc"}
                if sort_field:
                    sort_column = _get_column_by_field(cls, sort_field)
                    if sort_column is not None:
                        if _order and str(_order).strip().lower() in ['desc', 'descend', 'descending']:  # {"field": "name", "order": "desc"}
                            orders.append(desc(sort_column))
                        else:
                            orders.append(asc(sort_column))  # {"field": "name"}.
    return orders


def _get_column_by_field(cls, field, include_relationship=True):
    """
    Get the column of the specified class according to the field
    if not found in class and include_relationship==True, return the column of the relationship class
    """
    sort_column = cls.get_column_by_field(field)
    if include_relationship is True and sort_column is None:
        keys = field.split('.')
        if len(keys) == 2:
            relationships = cls.get_relationships()
            relationship_key = keys[0]
            relationship_filed = keys[1]
            relationship = relationships.get(relationship_key)
            if relationship:
                relationship_cls = relationship.mapper.class_
                sort_column = relationship_cls.get_column_by_field(relationship_filed)
    return sort_column


from ..utils import is_str, is_dict, is_list
from ..utils._private import get_dict, contains_any



import json

from flask import request

from .. import res_status_codes
from ..log import flaskz_logger, get_log_data
from ..models import model_to_dict, query_all_models, ModelMixin, parse_pss
from ..utils import get_list, is_dict, create_response, get_request_json


def init_model_rest_blueprint(model_cls, api_blueprint, url_prefix, module, routers=None, to_json_option=None, multiple_option=None):
    """
    ** Deprecated, please use register_model_route **

    Append rest api route to the api blueprint for the specified model class.

    Example:
        init_model_rest_blueprint(User, sys_mgmt_bp, '/user', 'user', multiple_option={
            'users': User,
            'roles': {
                'model_cls': Role,
                'option': {
                    'include': ['id', 'name']
                }
            }
        })

    :param to_json_option: The option used to return json, fields/cascade. ex){'cascade': 1}
    :param url_prefix: Api url path,ex) '/roles'
    :param module: The module name,ex) 'roles'
    :param api_blueprint:
    :param model_cls: The specified model class.
    :param routers: The api router list, ex)['query', 'query_pss', 'add', 'update', 'delete']
    :param multiple_option:

    :return: api_blueprint
    """
    if to_json_option is None:
        to_json_option = {}

    routers = get_list(routers, ['query', 'query_pss', 'query_multiple', 'add', 'update', 'upsert', 'delete'])

    if 'add' in routers:
        register_model_add_route(api_blueprint, model_cls, url_prefix, module, to_json_option=to_json_option)

    if 'delete' in routers:
        register_model_delete_route(api_blueprint, model_cls, url_prefix, module, to_json_option=to_json_option)

    if 'update' in routers:
        register_model_update_route(api_blueprint, model_cls, url_prefix, module, to_json_option=to_json_option)

    # if 'update' in routers:
    #     @api_blueprint.route(url_prefix + '/', methods=['PUT'])
    #     @rest_permission_required(module, 'update')
    #     @gen_route_method('update_put', url_prefix)
    #     def update():
    #         """
    #         todo
    #         Update the model data by the json in put mode.
    #         The difference between put and patch is that patch is partial update, and put is full update.            """
    #         pass

    if 'upsert' in routers:
        register_model_upsert_route(api_blueprint, model_cls, url_prefix, module, to_json_option=to_json_option)

    if 'query' in routers:
        register_model_query_route(api_blueprint, model_cls, url_prefix, module, to_json_option=to_json_option)

    if 'query_pss' in routers:
        register_model_query_pss_route(api_blueprint, model_cls, url_prefix, module, to_json_option=to_json_option, rule_suffix='query_pss')

    if multiple_option and 'query_multiple' in routers:
        register_models_query_route(api_blueprint, multiple_option, url_prefix, module, rule_suffix='query_multiple')

    return api_blueprint


def register_model_route(app, model, rule, module=None, types=None, multi_models=None, to_json_option=None, get_pss_config=None, strict_slash=True):
    """
    Register url rules for the specified model class to the application/blueprint.

    .. versionadded:: 1.5
    .. versionupdated::
        1.6.4 - add get_pss_config parameter

    Example:
        register_model_route(api_bp, User, 'users', 'users')

    :param app: Flask application / Blueprint instance
    :param model: The DB model class, ex)User
    :param rule: The URL rule string, ex)/users
    :param module: The module name for permission check, ex)users
    :param types: The types of the route, default is ['query', 'pss', 'multi', 'add', 'update', 'upsert', 'delete']
    :param multi_models: The multiple DB model classes
    :param to_json_option: The option to return the json, ex){'cascade': 1}
    :param get_pss_config: The callback function to custom pss query config
    :param strict_slash: If not false, the rule url will end with slash
    :return:
    """
    if type(get_pss_config) is bool:  # version match
        strict_slash = get_pss_config
        get_pss_config = None

    types = get_list(types, ['query', 'pss', 'multi', 'add', 'update', 'upsert', 'delete'])
    if 'add' in types:
        register_model_add_route(app, model, rule, module, to_json_option=to_json_option, strict_slash=strict_slash)
    if 'delete' in types:
        register_model_delete_route(app, model, rule, module, to_json_option=to_json_option, strict_slash=strict_slash)
    if 'update' in types:
        register_model_update_route(app, model, rule, module, to_json_option=to_json_option, strict_slash=strict_slash)
    if 'upsert' in types:
        register_model_upsert_route(app, model, rule, module, to_json_option=to_json_option, strict_slash=strict_slash)
    if 'query' in types:
        register_model_query_route(app, model, rule, module, to_json_option=to_json_option, strict_slash=strict_slash)
    if 'pss' in types:
        register_model_query_pss_route(app, model, rule, module, to_json_option=to_json_option, strict_slash=strict_slash, get_pss_config=get_pss_config)
    if multi_models and ('multi' in types or 'multiple' in types):
        register_models_query_route(app, multi_models, rule, module)
    if 'bulk_add' in types:
        register_model_bulk_add_route(app, model, rule, module, strict_slash=strict_slash)
    if 'bulk_delete' in types:
        register_model_bulk_delete_route(app, model, rule, module, strict_slash=strict_slash)
    if 'bulk_update' in types:
        register_model_bulk_update_route(app, model, rule, module, strict_slash=strict_slash)
    return app


def register_model_bulk_route(app, model, rule, module=None, types=None, with_relationship=False, strict_slash=True):
    """
    Register url rules for bulk operations of the specified model class to the application/blueprint.

    .. versionadded:: 1.8
    .. versionupdated::
        1.8.1   - add with_relationship param

    Example:
        register_model_bulk_route(api_bp, User, 'users', 'users')

    :param app: Flask application / Blueprint instance
    :param model: The DB model class, ex)User
    :param rule: The URL rule string, ex)/users
    :param module: The module name for permission check, ex)users
    :param types: The types of the route, default is ['bulk_add', 'bulk_delete', 'bulk_update']
    :param with_relationship: If True, the relationship data will be added/updated in bulk, default is False
    :param strict_slash: If not False, the rule url will end with slash
    :return:
    """
    types = get_list(types, ['bulk_add', 'bulk_delete', 'bulk_update'])
    if 'bulk_add' in types:
        register_model_bulk_add_route(app, model, rule, module, with_relationship=with_relationship, strict_slash=strict_slash)
    if 'bulk_delete' in types:
        register_model_bulk_delete_route(app, model, rule, module, strict_slash=strict_slash)
    if 'bulk_update' in types:
        register_model_bulk_update_route(app, model, rule, module, with_relationship=with_relationship, strict_slash=strict_slash)
    return app


def register_model_add_route(app, model, rule, module=None, action='add', methods=None, to_json_option=None, strict_slash=True, endpoint=None):
    """
    Register a add type URL rule for the specified model class to the application/blueprint.

    Examples:
        register_model_add_route(api_blueprint, User, 'users', 'users')

    :param app: Flask application / Blueprints instance
    :param model: The DB model class, ex)User
    :param rule: The URL rule string, ex)/users
    :param module: The module name for permission check, ex)users
    :param action: The action of the module for permission check, ex)add
    :param methods: The methods of the route, default is ['POST']
    :param to_json_option: The option to return the json, ex){'cascade': 1}
    :param strict_slash: If not false, the rule url will end with slash
    :param endpoint: The name of the route endpoint, default is None(use view function name as endpoint name)
    :return:
    """
    methods = methods or ['POST']
    base_rule, did_rule, suffix_rule = _gen_route_rule(rule, strict_slash=strict_slash)

    @app.route(base_rule, methods=methods, endpoint=endpoint)
    @rest_permission_required(module, action)
    @gen_route_method('add', base_rule)
    def add():
        request_json = request.json
        req_log_data = json.dumps(request_json)

        success, data = model.add(request_json)
        res_data = model_to_dict(data, to_json_option)

        res_log_data = get_log_data(res_data)
        log_operation(module, 'add', success, req_log_data, res_log_data)
        flaskz_logger.info(get_rest_log_msg('Add {} data'.format(model.get_class_name()), req_log_data, success, res_log_data))

        return create_response(success, res_data)


def register_model_delete_route(app, model, rule, module=None, action='delete', methods=None, to_json_option=None, strict_slash=True, endpoint=None):
    """
    Register a delete type URL rule for the specified model class to the application/blueprint.

    .. versionupdated::
        1.2   - add path converter of the did
        1.3.1 - remove path converter of the did(Flask<2.2.3 path converter doesn't redirect to trailing slash)

    Examples:
        register_model_delete_route(api_blueprint, User, 'users', 'users')

    :param app: Flask application / Blueprints instance
    :param model: The DB model class, ex)User
    :param rule: The URL rule string, ex)/users
    :param module: The module name for permission check, ex)users
    :param action: The action of the module for permission check, ex)delete
    :param methods: The methods of the route, default is ['DELETE']
    :param to_json_option: The option to return the json, ex){'cascade': 1}
    :param strict_slash: If not false, the rule url will end with slash
    :param endpoint: The name of the route endpoint, default is None(use view function name as endpoint name)
    :return:
    """
    methods = methods or ['DELETE']
    base_rule, did_rule, suffix_rule = _gen_route_rule(rule, strict_slash=strict_slash, did_suffix='<did>')

    @app.route(did_rule, methods=methods, endpoint=endpoint)
    @rest_permission_required(module, action)
    @gen_route_method('delete', base_rule)
    def delete(did):
        success, data = model.delete(did)
        res_data = model_to_dict(data, to_json_option)

        res_log_data = get_log_data(res_data)
        log_operation(module, 'delete', success, did, res_log_data)
        flaskz_logger.info(get_rest_log_msg('Delete {} data'.format(model.get_class_name()), did, success, res_log_data))

        return create_response(success, res_data)


def register_model_update_route(app, model, rule, module=None, action='update', methods=None, to_json_option=None, strict_slash=True, endpoint=None):
    """
    Register update type URL rule for the specified model class to the application/blueprint.
    The priority of the primary key in the url is higher than that in the body.

    Examples:
        register_model_update_route(api_blueprint, User, 'users', 'users')

    :param app: Flask application / Blueprints instance
    :param model: The DB model class, ex)User
    :param rule: The URL rule string, ex)/users
    :param module: The module name for permission check, ex)users
    :param action: The action of the module for permission check, ex)update
    :param methods: The methods of the route, default is ['PATCH']
    :param to_json_option: The option to return the json, ex){'cascade': 1}
    :param strict_slash: If not false, the rule url will end with slash
    :param endpoint: The name of the route endpoint, default is None(use view function name as endpoint name)
    :return:
    """
    methods = methods or ['PATCH']
    base_rule, did_rule, suffix_rule = _gen_route_rule(rule, strict_slash=strict_slash, did_suffix='<did>')

    @app.route(base_rule, methods=methods, endpoint=endpoint)
    @app.route(did_rule, methods=methods, endpoint=endpoint)
    @rest_permission_required(module, action)
    @gen_route_method('update', base_rule)
    def update(did=None):
        request_json = request.json
        if did is not None:
            request_json[model.get_primary_field()] = did  # use pk in url
        req_log_data = json.dumps(request_json)

        success, data = model.update(request_json)
        res_data = model_to_dict(data, to_json_option)

        res_log_data = get_log_data(res_data)
        log_operation(module, 'update', success, req_log_data, res_log_data)
        flaskz_logger.info(get_rest_log_msg('Update {} data'.format(model.get_class_name()), req_log_data, success, res_log_data))

        return create_response(success, res_data)


def register_model_upsert_route(app, model, rule, module=None, action='upsert', methods=None, to_json_option=None, strict_slash=True, rule_suffix='upsert', endpoint=None):
    """
    Register a upsert type URL rule for the specified model class to the application/blueprint.
    If primary key value is in json, perform an update action, otherwise perform an add action.

    Examples:
        register_model_upsert_route(api_blueprint, User, 'users', 'users')

    :param app: Flask application / Blueprints instance
    :param model: The DB model class, ex)User
    :param rule: The URL rule string, ex)/users
    :param module: The module name for permission check, ex)users
    :param action: The action of the module for permission check, ex)update
    :param methods: The methods of the route, default is ['PATCH']
    :param to_json_option: The option to return the json, ex){'cascade': 1}
    :param strict_slash: If not false, the rule url will end with slash
    :param rule_suffix: The upsert suffix, default is 'upsert'
    :param endpoint: The name of the route endpoint, default is None(use view function name as endpoint name)
    """
    methods = methods or ['POST']
    base_rule, did_rule, suffix_rule = _gen_route_rule(rule, strict_slash=strict_slash, rule_suffix=rule_suffix)

    @app.route(suffix_rule, methods=methods, endpoint=endpoint)
    @rest_permission_required(module, action)
    @gen_route_method('upsert', base_rule)
    def upsert():
        request_json = request.json
        req_log_data = json.dumps(request_json)

        if request_json.get(model.get_primary_field()):
            upsert_action = "update"
            success, data = model.update(request_json)
        else:
            upsert_action = "add"
            success, data = model.add(request_json)

        res_data = model_to_dict(data, to_json_option)
        res_log_data = get_log_data(res_data)
        log_operation(module, upsert_action, success, req_log_data, res_log_data)
        flaskz_logger.info(get_rest_log_msg('Upsert {} data'.format(model.get_class_name()), req_log_data, success, res_log_data))

        return create_response(success, res_data)


def register_model_query_route(app, model, rule, module=None, action=None, methods=None, to_json_option=None, strict_slash=True, endpoint=None):
    """
    Register query type URL rule for the specified model class to the application/blueprint.

    Examples:
        register_model_query_route(api_blueprint, User, 'users', 'users')

    :param app: Flask application / Blueprints instance
    :param model: The DB model class, ex)User
    :param rule: The URL rule string, ex)/users
    :param module: The module name for permission check, ex)users
    :param action: The action of the module for permission check, default is None(only check module permission)
    :param methods: The methods of the route, default is ['GET']
    :param to_json_option: The option to return the json, ex){'cascade': 1}
    :param strict_slash: If not false, the rule url will end with slash
    :param endpoint: The name of the route endpoint, default is None(use view function name as endpoint name)
    :return:
    """
    methods = methods or ['GET']
    base_rule, did_rule, suffix_rule = _gen_route_rule(rule, strict_slash=strict_slash, did_suffix='<did>')

    @app.route(base_rule, methods=methods, endpoint=endpoint)
    @app.route(did_rule, methods=methods, endpoint=endpoint)
    @rest_permission_required(module, action)
    @gen_route_method('query', base_rule)
    def query(did=None):
        if did is None:
            success, data = model.query_all()
        else:
            try:
                data = model.query_by_pk(did)
                if data:
                    success, data = True, data
                else:
                    success, data = False, res_status_codes.db_data_not_found
            except Exception as e:
                flaskz_logger.exception(e)
                success, data = False, res_status_codes.db_query_err

        res_data = model_to_dict(data, to_json_option)
        flaskz_logger.debug(get_rest_log_msg('Query {} data'.format(model.get_class_name()), did, success, res_data))
        return create_response(success, res_data)


def register_model_query_pss_route(app, model, rule, module=None, action=None, methods=None, to_json_option=None, strict_slash=True, rule_suffix='pss', get_pss_config=None,
                                   endpoint=None):
    """
    Register pss query URL rule for the specified model class to the application/blueprint.
    pss = paging + search + sort

    .. versionadded::
        1.6.4 - add get_pss_config param

    Examples:
        register_model_query_pss_route(api_blueprint, User, 'users', 'users', get_pss_config=append_search_config)

    :param app: Flask application / Blueprints instance
    :param model: The DB model class, ex)User
    :param rule: The URL rule string, ex)/users
    :param module: The module name for permission check, ex)users
    :param action: The action of the module for permission check, default is None(only check module permission)
    :param methods: The methods of the route, default is ['GET', 'POST']
    :param to_json_option: The option to return the json, ex){'cascade': 1}
    :param strict_slash: If not false, the rule url will end with slash
    :param rule_suffix: The pss suffix, default is 'query_pss'
    :param get_pss_config: Callback function used to generate pss config.
    :param endpoint: The name of the route endpoint, default is None(use view function name as endpoint name)
    :return:
    """
    methods = methods or ['GET', 'POST']
    base_rule, did_rule, suffix_rule = _gen_route_rule(rule, strict_slash=strict_slash, rule_suffix=rule_suffix)

    @app.route(suffix_rule, methods=methods, endpoint=endpoint)
    @rest_permission_required(module, action)
    @gen_route_method('query_pss', base_rule)
    def query_pss():
        request_json = get_request_json({})  # @2023-06-15, request.json --> get_request_json({})
        req_log_data = json.dumps(request_json)

        if callable(get_pss_config):  # @2023-10-13 add
            request_json = get_pss_config(request_json)  # @2023-10-23 fix, get_pss_config-->request_json

        success, data = model.query_pss(parse_pss(model, request_json))
        if success is True:
            data['data'] = model_to_dict(data.get('data', []), to_json_option)

        flaskz_logger.debug(get_rest_log_msg('Query pss {} data'.format(model.get_class_name()), req_log_data, success, data))
        return create_response(success, data)


def register_models_query_route(app, models, rule, module=None, action=None, methods=None, strict_slash=True, rule_suffix='multi', endpoint=None):
    """
    Register query multi models URL rule to the application/blueprint.
    pss = paging + search + sort

    Examples:
        register_models_query_route(api_blueprint,
                                 {
                                     'users': User,
                                     'roles': {
                                         'model_cls': Role,
                                         'option': {
                                             # 'cascade': 1,
                                             'include': ['id', 'name']
                                             # 'filter': lambda ins: ins.type == 'local'
                                         }
                                     }
                                 })

    :param app: Flask application / Blueprints instance
    :param models: The multiple DB model classes,
    :param rule: The URL rule string, ex)/users
    :param module: The module name for permission check, ex)users
    :param action: The action of the module for permission check, default is None(only check module permission)
    :param methods: The methods of the route, default is ['GET', 'POST']
    :param strict_slash: If not false, the rule url will end with slash
    :param rule_suffix: The pss suffix, default is 'multi'
    :param endpoint: The name of the route endpoint, default is None(use view function name as endpoint name)

    :return:
    """
    methods = methods or ['GET']
    base_rule, did_rule, suffix_rule = _gen_route_rule(rule, strict_slash=strict_slash, rule_suffix=rule_suffix)

    @app.route(suffix_rule, methods=methods, endpoint=endpoint)
    @rest_permission_required(module, action)
    @gen_route_method('query_multi', base_rule)
    def query_multi():
        model_cls_list = []
        multi_list = []
        for key in models:
            item = models[key]
            item_option = {}
            m_cls = None
            if is_dict(item):  # 'roles': {'model': Role,'option': {'include': ['id', 'name']}}
                m_cls = item.get('model_cls') or item.get('model')
                item_option = item.get('option') or item.get('to_json_option')
            elif isinstance(item, type) and issubclass(item, ModelMixin):  # 'users': User,
                m_cls = item
                item_option = {}
            if m_cls and issubclass(m_cls, ModelMixin):
                model_cls_list.append(m_cls)
                multi_list.append({
                    'field': key,
                    'option': item_option
                })
        result = query_all_models(*model_cls_list)

        if type(result) is tuple:
            success = False
            res_data = model_to_dict(result[1])
        else:
            success = True
            res_data = {}
            for index, item in enumerate(multi_list):
                res_data[item.get('field')] = model_to_dict(result[index], item.get('option'))

        flaskz_logger.debug(get_rest_log_msg('Query multi {} data'.format(model_cls_list), None, success, res_data))
        return create_response(success, res_data)


def register_model_bulk_add_route(app, model, rule, module=None, action='add', methods=None, with_relationship=False, strict_slash=True, rule_suffix='bulk', endpoint=None):
    """
    Register bulk add type URL rule for the specified model class to the application/blueprint.

    .. versionadded:: 1.8
    .. versionupdated::
        1.8.1   - add with_relationship param

    Examples:
        register_model_bulk_add_route(api_blueprint, User, 'users', 'users')

    :param app: Flask application / Blueprints instance
    :param model: The DB model class, ex)User
    :param rule: The URL rule string, ex)/users
    :param module: The module name for permission check, ex)users
    :param action: The action of the module for permission check, ex)add
    :param methods: The methods of the route, default is ['POST']
    :param with_relationship: If True, the relationship data will be added in bulk, default is False
    :param strict_slash: If not False, the rule url will end with slash
    :param rule_suffix: The upsert suffix, default is 'bulk-add'
    :param endpoint: The name of the route endpoint, default is None(use view function name as endpoint name)
    :return:
    """
    methods = methods or ['POST']
    base_rule, did_rule, suffix_rule = _gen_route_rule(rule, strict_slash=strict_slash, rule_suffix=rule_suffix)

    @app.route(suffix_rule, methods=methods, endpoint=endpoint)
    @rest_permission_required(module, action)
    @gen_route_method('bulk_add', base_rule)
    def bulk_add():
        request_json = request.json
        req_log_data = json.dumps(request_json)
        try:
            model.bulk_add(request_json, with_relationship=with_relationship)
            success, res_data = True, request_json
        except Exception as e:
            flaskz_logger.exception(e)
            success, res_data = False, res_status_codes.db_add_err

        res_log_data = get_log_data(res_data)
        log_operation(module, 'add', success, req_log_data, res_log_data)
        flaskz_logger.info(get_rest_log_msg('Bulk add {} data'.format(model.get_class_name()), req_log_data, success, res_log_data))

        return create_response(success, res_data)


def register_model_bulk_delete_route(app, model, rule, module=None, action='delete', methods=None, strict_slash=True, rule_suffix='bulk', endpoint=None):
    """
    Register bulk delete type URL rule for the specified model class to the application/blueprint.

    .. versionadded:: 1.8

    Examples:
        register_model_bulk_delete_route(api_blueprint, User, 'users', 'users')

    :param app: Flask application / Blueprints instance
    :param model: The DB model class, ex)User
    :param rule: The URL rule string, ex)/users
    :param module: The module name for permission check, ex)users
    :param action: The action of the module for permission check, ex)delete
    :param methods: The methods of the route, default is ['DELETE']
    :param strict_slash: If not false, the rule url will end with slash
    :param rule_suffix: The upsert suffix, default is 'bulk-delete'
    :param endpoint: The name of the route endpoint, default is None(use view function name as endpoint name)
    :return:
    """
    methods = methods or ['DELETE']
    base_rule, did_rule, suffix_rule = _gen_route_rule(rule, strict_slash=strict_slash, did_suffix='<ids>', rule_suffix=rule_suffix)

    @app.route(suffix_rule, methods=methods, endpoint=endpoint)
    @app.route(did_rule, methods=methods, endpoint=endpoint)
    @rest_permission_required(module, action)
    @gen_route_method('bulk_delete', base_rule)
    def bulk_delete(ids=None):
        req_log_data = ids
        if ids is None:
            request_json = request.json
            req_log_data = json.dumps(request_json)
            ids = request_json
            if type(request_json) is dict:
                ids = request_json.get('ids') or request_json.get('id', [])
        else:
            ids = ids.split(',')
        try:
            success, res_data = True, model.bulk_delete(ids)
        except Exception as e:
            flaskz_logger.exception(e)
            success, res_data = False, res_status_codes.db_delete_err

        res_log_data = get_log_data(res_data)
        log_operation(module, 'delete', success, req_log_data, res_log_data)
        flaskz_logger.info(get_rest_log_msg('Bulk delete {} data'.format(model.get_class_name()), req_log_data, success, res_log_data))

        return create_response(success, res_data)


def register_model_bulk_update_route(app, model, rule, module=None, action='update', methods=None, with_relationship=False, strict_slash=True, rule_suffix='bulk', endpoint=None):
    """
    Register bulk update type URL rule for the specified model class to the application/blueprint.

    .. versionadded:: 1.8
    .. versionupdated::
        1.8.1   - add with_relationship param

    Examples:
        register_model_bulk_update_route(api_blueprint, User, 'users', 'users')

    :param app: Flask application / Blueprints instance
    :param model: The DB model class, ex)User
    :param rule: The URL rule string, ex)/users
    :param module: The module name for permission check, ex)users
    :param action: The action of the module for permission check, ex)update
    :param methods: The methods of the route, default is ['PATCH']
    :param with_relationship: If True, the relationship data will be updated in bulk, default is False
    :param strict_slash: If not False, the rule url will end with slash
    :param rule_suffix: The upsert suffix, default is 'bulk-update'
    :param endpoint: The name of the route endpoint, default is None(use view function name as endpoint name)
    :return:
    """
    methods = methods or ['PATCH']
    base_rule, did_rule, suffix_rule = _gen_route_rule(rule, strict_slash=strict_slash, rule_suffix=rule_suffix)

    @app.route(suffix_rule, methods=methods, endpoint=endpoint)
    @rest_permission_required(module, action)
    @gen_route_method('bulk_update', base_rule)
    def bulk_update():
        request_json = request.json
        req_log_data = json.dumps(request_json)
        try:
            model.bulk_update(request_json, with_relationship=with_relationship)
            success, res_data = True, request_json
        except Exception as e:
            flaskz_logger.exception(e)
            success, res_data = False, res_status_codes.db_update_err

        res_log_data = get_log_data(res_data)
        log_operation(module, 'update', success, req_log_data, res_log_data)
        flaskz_logger.info(get_rest_log_msg('Bulk update {} data'.format(model.get_class_name()), req_log_data, success, res_log_data))

        return create_response(success, res_data)


def _gen_route_rule(rule, *, strict_slash=True, did_suffix=None, rule_suffix=None):
    if not rule.startswith('/'):
        rule = '/' + rule
    base_rule = rule
    suffix_rule = rule
    did_rule = None

    if rule_suffix is not None:
        suffix_rule = rule = rule.rstrip('/') + '/' + rule_suffix

    if did_suffix is not None:
        did_rule = rule.rstrip('/') + '/' + did_suffix

    if strict_slash is not False:
        base_rule = base_rule.rstrip('/') + '/'
        suffix_rule = suffix_rule.rstrip('/') + '/'
        did_rule = did_rule.rstrip('/') + '/' if did_rule else None

    return base_rule, did_rule, suffix_rule


from ._mgmt import *
from ._util import *



import warnings

from flask import request

__all__ = ['get_remote_addr', 'is_ajax', 'get_request_json', 'get_pss']


def get_remote_addr():
    """
    Get the remote ip address of the current request.

    :return:
    """
    if request:
        return request.environ.get('HTTP_X_REAL_IP', request.remote_addr)


def is_ajax():
    """
    Check if the request is an ajax request.

    :return:
    """
    if request:
        return request.headers.get('X-Requested-With') == 'XMLHttpRequest'
    return False


def get_request_json(*args):
    """
    Get the JSON data(parsed) in request.
    If json does not exist or parsing json error, return {}

    .. versionadded:: 1.3

    :return:
    """
    data = None
    try:
        data = request.get_json(force=True, silent=True)
    except Exception:
        pass
    if data is None:
        if len(args) > 0:
            return args[0]
    return data


def get_pss(*args, **kwargs):
    warnings.warn('flaskz.utils.get_pss() has been replaced by flaskz.models.parse_pss()', category=DeprecationWarning)
    return parse_pss(*args, **kwargs)


from ..models._query_util import parse_pss



import inspect
import textwrap
from urllib import parse as urllib_parse

from flask import request as flask_request
from requests import Session, sessions
from requests.adapters import HTTPAdapter

from ._common import is_dict
from .. import res_status_codes
from ..log import flaskz_logger

__all__ = ['request', 'api_request', 'forward_request', 'append_url_search_params']

_session_request_kwargs = None


def api_request(url, method="GET", url_params=None, base_url="", raw_response=False, **kwargs):
    """
    Request an api.

   .. versionupdated::
        1.7.2 - add `http_kwargs` param to set http request parameters(HTTPAdapter)

    Example
        # post request
        api_request('req_url', 'POST', json=payload)
        api_request({'url': 'req_url', 'method': 'POST'}, json=payload)

        # get request
        api_request('req_url', headers={'Authorization': 'eyJhbGciOiJIUzUxMiIs......'})

        # base url
        api_request('req_url', base_url='base_url') # url --> 'base_url/req_url'

        # url_variables
        api_request('req_url/{id}/', url_variables={'id': 1}) # url --> 'req_url/1/'

        # url_search_params
        api_request('req_url', url_search_params={'id': 1}) # url --> 'req_url?id=1'

        # http_kwargs
        api_request('req_url', http_kwargs={ # --> HTTPAdapter parameters
            'pool_connections': 10, 'pool_maxsize': 100, 'max_retries': 10
        })

    :param url:
    :param method:
    :param url_params:
    :param raw_response:
    :param base_url:
    :param kwargs:
    :return:
    """
    # warnings.warn('flaskz.utils.api_request() has been replaced by flaskz.utils.request()', category=DeprecationWarning)

    success, result = request(url, method=method, url_params=url_params, base_url=base_url, raw_response=raw_response, **kwargs)
    if success:
        return result
    return res_status_codes.api_request_err, result


def request(url, method="GET", url_params=None, base_url="", raw_response=False, **kwargs):
    """
    Request an api.

   .. versionadded:: 1.8.0 - add to replace api_request @2024-05-02

    Example
        # post request
        request('req_url', 'POST', json=payload)
        request({'url': 'req_url', 'method': 'POST'}, json=payload)

        # get request
        request('req_url', headers={'Authorization': 'eyJhbGciOiJIUzUxMiIs......'})

        # base url
        request('req_url', base_url='base_url') # url --> 'base_url/req_url'

        # url_variables
        request('req_url/{id}/', url_variables={'id': 1}) # url --> 'req_url/1/'

        # url_search_params
        request('req_url', url_search_params={'id': 1}) # url --> 'req_url?id=1'

        # http_kwargs
        request('req_url', http_kwargs={ # --> HTTPAdapter parameters
            'pool_connections': 10, 'pool_maxsize': 100, 'max_retries': 10
        })

    :param url:
    :param method:
    :param url_params:
    :param raw_response:
    :param base_url:
    :param kwargs:
    :return:
    """
    _method = method
    if is_dict(url):
        _url = url.get('url')
        if 'method' in url:
            _method = url.get("method")
    else:
        _url = url

    if base_url:
        base_url = base_url.rstrip('/')
        _url = _url.lstrip('/')
        _url = base_url + '/' + _url

    if is_dict(url_params):  # @2023-05-30 update variables(for replace)
        _url = _url.format(**url_params)

    url_variables = kwargs.get('url_variables', None)  # @2023-05-30 add variables(for replace)
    if is_dict(url_variables):
        _url = _url.format(**url_variables)

    url_search_params = kwargs.get('url_search_params', None)  # @2023-05-30 add search_params(for append)
    if url_search_params is not None:
        _url = append_url_search_params(_url, url_search_params)

    _method = _method.strip().upper()

    flaskz_logger.debug('Api request:\n   url={url}\n   method={method}\n   data={data}\n   json={json}'.format(**{
        'url': _url,
        'method': _method,
        'data': kwargs.get('data'),
        'json': kwargs.get('json')
    }))
    try:
        res = _request(method=_method, url=_url, **kwargs)  # @2024-01-16 change, requests.request-->_request
        status_code = res.status_code
        result = res.text
        flaskz_logger.debug('Api request completed:\n   status_code={status_code}\n   result={result}'.format(**{
            'status_code': status_code,
            # 'result': slice_str(result, 60, 10, '\n......\n'),
            'result': textwrap.shorten(result, 1000)  # @2022-05-26:将日志输出由slice_str改为了shorten
        }))
        if kwargs.get('close') is not False:  # @2024-01-09 add
            res.close()
        if raw_response is True:
            return True, res
        return True, result
    except Exception as e:
        result = str(e)
        flaskz_logger.exception('Api request failed:\n' + result)

    return False, result


def forward_request(url, payload=None, raw_response=False, error_code=500, **kwargs):
    """
    Forward the request to other service.

    Example:
        @api_bp.route('/<regex(".*"):path>/', methods=HTTP_METHODS)
        def remote(path):
            remote_res = forward_request(base_url + path)
            res = make_response(remote_res[0], remote_res[1])
            for k, v in remote_res[2]:
                if k not in ['Transfer-Encoding']:
                    res.headers[k] = v
            return res

    :param error_code:
    :param payload:
    :param raw_response:
    :param url: The forward url
    :return:
    """
    req_kwargs = {'url': url}
    _payload = payload or ['method', 'data', 'json', 'headers', 'cookies']
    for item in _payload:
        if item == 'json':  # @2022-05-06: fix request.json -->BadRequest('Content-Type was not 'application/json')
            req_kwargs[item] = flask_request.get_json(force=True, silent=True)
        else:
            req_kwargs[item] = getattr(flask_request, item, None)
    req_kwargs['params'] = flask_request.args  # @2022-06-25 add query string
    req_kwargs.update(kwargs)  # kwargs high priority

    url_params = req_kwargs.get('url_params')
    if url_params is None:  # if url_params is none, append request.view_args
        req_kwargs['url_params'] = flask_request.view_args or {}

    req_kwargs['raw_response'] = True

    success, res = request(**req_kwargs)
    if success is False:
        return res, error_code
    if raw_response is True:
        return res
    return res.text, res.status_code, res.headers.items()


def append_url_search_params(url, params):  # @2022-05-09: add
    """
    Appends specified key/value pair as search parameter.
    Only append the key if its value is None.

    Example:
        append_url_search_params('https://example.com',{'foo':1,'bar':2, 'xx': None}) # 'https://example.com?foo=1&bar=2&xx' # append
        append_url_search_params('https://example.com?foo=1&bar=2',{'baz':3}) # 'https://example.com?foo=1&bar=2&baz=3' # append
        append_url_search_params('https://example.com?foo=1&bar=2',{'bar':3}) # 'https://example.com?foo=1&bar=3' # replace
        append_url_search_params('a/b',{'c':3,'d':None}) # 'a/b?c=3&d'

        append_url_search_params('https://example.com', ['a', 'b', 'c=2']) # 'https://example.com?a&b&c=2'
        append_url_search_params('https://example.com', 'a=1&b=2&c')  # 'https://example.com?a=1&b=2&c'

    :param url:
    :param params:
    :return:

    """
    url_parts = urllib_parse.urlparse(url)
    query = dict(urllib_parse.parse_qsl(url_parts.query))

    kv_params = {}
    k_params = []
    params_type = type(params)
    if params_type is list:
        k_params = params
    elif params_type is str:
        k_params = [params]
    elif params_type is dict:
        for k, v in params.items():
            if v is not None:
                kv_params[k] = v
            else:
                k_params.append(k)
        query.update(kv_params)

    query_string = urllib_parse.urlencode(query)
    if len(k_params) > 0:
        k_str = '&'.join(k_params)
        if query_string:
            if not k_str.startswith('&'):
                k_str = '&' + k_str
            query_string += k_str
        else:
            query_string += k_str

    return url_parts._replace(query=query_string).geturl()


def _request(method, url, **kwargs):
    # requests.request
    # @2024-01-26 add
    with sessions.Session() as session:
        http_kwargs = kwargs.get('http_kwargs', None)
        if http_kwargs:
            adapter = HTTPAdapter(**http_kwargs)
            session.mount('http://', adapter)
            session.mount('https://', adapter)
        return session.request(method=method, url=url, **_get_session_request_kwargs(kwargs))


def _get_session_request_kwargs(kwargs):
    global _session_request_kwargs
    if _session_request_kwargs is None:
        _session_request_kwargs = []
        for key in inspect.signature(Session.request).parameters:
            _session_request_kwargs.append(key)

    req_kwargs = {}
    for key, value in kwargs.items():
        if key in _session_request_kwargs:
            req_kwargs[key] = value
    return req_kwargs



# response status for client use
# if value is tuple, the second value will be returned as the error message

# database
db_add_err = 'db_add_err', 'Database Add Error'
db_delete_err = 'db_delete_err', 'Database Delete Error'
db_update_err = 'db_update_err', 'Database Update Error'
db_query_err = 'db_query_err', 'Database Query Error'

db_data_not_found = 'db_data_not_found', 'Data Not Found'
db_data_already_exist = 'db_data_already_exist', 'Data Already Exists'
db_data_in_use = 'db_data_in_use', 'Data In Use'

# remote request
api_request_err = "api_req_err", 'Remote api request error'

# error
uri_unauthorized = 'uri_unauthorized', 'Unauthorized'  # 401 Unauthorized, need login
uri_forbidden = 'uri_forbidden', 'Forbidden'  # 403 Forbidden, need permission
uri_not_found = 'uri_not_found', 'Not Found'  # 404 Not Found
method_not_allowed = 'method_not_allowed', 'Method Not Allowed'  # 405 HTTP method is not supported
internal_server_error = 'internal_server_error', 'Internal Server Error'  # 500
bad_request = 'bad_request', 'Bad Request'  # 400 Bad Request, payload error

# account
account_not_found = 'account_not_found', 'No Account Found'
account_disabled = 'account_disabled', "Account Disabled"
account_verify_err = 'account_verify_err', 'Wrong Password'

# others
